(*
 *  emlClass.sml: Class specific fuctions and information.
 *)

(* 
 *  Used to give each class a unique id.
 *  Initialized to 0.
 *)
val classIdCounter = ref 0;

(* 
 *  Increases the class id counter and returns the id.
 *)
fun getNextId() = 
    let val currentId = !classIdCounter
    in (classIdCounter := currentId + 1;
	currentId)
    end

fun getClassTypeOption(Class{name, id, mode, superclass, interfaces, dType}) =
        dType

fun getClassType(cl) =
    let
        val clTypeOption = getClassTypeOption(cl)
    in
        case (clTypeOption) of
            NONE => raise TypeError("Class does not have a type!")
          | SOME(clType) => clType
    end

fun firstIsChildOfSecond(Class{name, id, mode, superclass=NONE, interfaces,
        dType}, _) = false
  | firstIsChildOfSecond(Class{name, id, mode, superclass=SOME(Class{name=sn,
        id=superID, mode=sm, superclass=ss, interfaces=sis, dType=st}),
        interfaces, dType}, Class{name=pn, id=parentID, mode=pm, superclass=ps,
        interfaces=pis, dType=pt}) =
    superID = parentID

(* 
 *  Compares two classes to see if they are the same.  This is done by comparing
 *  the two class ids.  If they are the same, then the classes are the same.
 *)
fun areSameClass(Class{name=n1, id=id1, mode=m1, superclass=s1, interfaces=i1,
        dType=t1}, Class{name=n2, id=id2, mode=m2, superclass=s2,
        interfaces=i2, dType=t2}) = 
    id1 = id2

(* 
 *  Compares two classes to see if the first is a subclass of the second.
 *  This is done by recursively checking to see if the classes are the same.
 *  If the classes are the same, then the first is a subclass of the second.
 *  If they are not the same, check to see if the superclass of the first is
 *  a subclass of the second (by the same reasoning).  Repeat until a match is
 *  found (so the first IS a subclass of the second), or until the first does
 *  not have a superclass (first is NOT a subclass of the second). 
 *)
fun FirstIsSubclassOfSecond(cl1 as Class{name=n1, id=id1, mode=m1,
        superclass=s1, interfaces=i1, dType=t1}, cl2 as Class{name=n2, id=id2,
        mode=m2, superclass=s2, interfaces=i2, dType=t2}) = 
    if areSameClass(cl1, cl2) then true
    else 
        case s1 of
            NONE => false
          | SOME (superClass) => FirstIsSubclassOfSecond(superClass,cl2)

(* 
 *  Compare two classes to see if the first is a strict subclass of the second.
 *  This is done by making sure that the two classes are not the same, and then 
 *  calling FirstIsSubclassOfSecond() on the elements (this will allow the 
 *  recursive checking to occur).
 *)
fun FirstIsStrictSubclassOfSecond(class1, class2) = 
    if areSameClass(class1, class2) then false
    else FirstIsSubclassOfSecond(class1, class2)

(*
 *  Given one class and an interface, check to see if the class implements
 *  the given interface.
 *)
fun classImplementsInterface(Class{name=n1, id=id1, mode=m1, superclass=s1,
        interfaces=i1, dType=t1}, interface) =
    let
        fun checkInterfaceList(nil) = false
          | checkInterfaceList(curInterface::restInterfaces) =
            if (areSameClass(curInterface, interface)) then
                true
            else
                checkInterfaceList(restInterfaces)
    in
        checkInterfaceList(i1)
    end

(*
 *  Given a name environment, we look through it and return all the direct
 *  children of the given class.
 *)
fun getClassChildren(parentClass, name_env) =
    let
        fun checkChild((name:string, value:NenvValue), curChildList):Class list =
            case (value) of
                ntNotAClass => curChildList
              | ntClass(potentialChild) => 
                    if (firstIsChildOfSecond(potentialChild, parentClass)) then
                        potentialChild :: curChildList
                    else
                        curChildList
    in
        Assoc_List.iterate(checkChild, nil:Class list, name_env)
    end

(*
 *  Given a name environment, we look through it and return all classes that
 *  implement the given class.
 *)
fun getImplementingClasses(interface, name_env) =
    let
        fun checkForInterface((name:string, value:NenvValue),
                curImplementingList):Class list =
            case (value) of
                ntNotAClass => curImplementingList
              | ntClass(potentialImplementor) => 
                    if (classImplementsInterface(potentialImplementor,
                            interface)) then
                        potentialImplementor :: curImplementingList
                    else
                        curImplementingList
    in
        Assoc_List.iterate(checkForInterface, nil:Class list, name_env)
    end

fun FirstIsSubtypeOfSecond(cl1, cl2 as Class{name=n2, id=id2,
        mode=Interface, superclass=s2, interfaces=i2, dType=t2}) = 
    if classImplementsInterface(cl1, cl2) then
        true
    else
        false
(*
 *  If we're not comparing against an interface then we just use the original
 *  check (check to see if it is a subclass.
 *)
  | FirstIsSubtypeOfSecond(cl1, cl2) = FirstIsSubclassOfSecond(cl1, cl2)
