(* 
 *  emlEnv.sml: Environment functions and interactions.
 *)

(* 
 *  An environment stores a list of bindings, one for each variable
 *  that's bound, mapping each variable to its current Value.
 *)
type Environment = (string, Value) Assoc_List.assoc_list
type NEnvironment = (string, NenvValue) Assoc_List.assoc_list
type tEnvironment = (string, Type) Assoc_List.assoc_list
type itcEnvironment = {name:string, gblType:Type, pats:PatternIntern list} list

(* 
 *  empty_environment is the initial environment with no bound variables.
 *)
val empty_environment:Environment = Assoc_List.empty
val empty_nenvironment:NEnvironment = Assoc_List.empty
val empty_tenvironment:tEnvironment = Assoc_List.empty
val empty_itcenvironment:itcEnvironment = nil

(* 
 *  add_binding adds a name->value binding to the environment,
 *  returning a new extended environment.  add_binding probably doesn't
 *  need to be called directly from your solution. 
 *)
fun add_binding(env:Environment, name:string, value:Value):Environment =
    Assoc_List.store(env, name, value)

fun add_nbinding(nenv:NEnvironment, name:string, value:NenvValue):NEnvironment =
    Assoc_List.store(nenv, name, value)

fun add_tbinding(tenv:tEnvironment, name:string, value:Type):tEnvironment =
    Assoc_List.store(tenv, name, value)

(* 
 *  new_binding creates a new environment containing a single binding
 *  of name->value.
 *)
fun new_binding(name:string, value:Value):Environment =
    add_binding(empty_environment, name, value)

fun new_nbinding(name:string, value:NenvValue):NEnvironment = 
    add_nbinding(empty_nenvironment, name, value)

fun new_tbinding(name:string, value:Type):tEnvironment = 
    add_tbinding(empty_tenvironment, name, value)

(* 
 *  combine_envs takes two environments and adds them together to form
 *  a new, merged environment containing all the bindings of the two
 *  combined environments.  if any name is defined in both enviroments,
 *  the binding in env2 takes precedence (shadows) that in env1.
 *)
fun combine_envs(env1:Environment, env2:Environment):Environment =
    let fun combine_one((name:string, value:Value), env):Environment =
	        add_binding(env, name, value)
    in Assoc_List.iterate(combine_one, env1, env2) end

fun combine_nenvs(nenv1:NEnvironment, nenv2:NEnvironment):NEnvironment = 
    let fun combine_one((name:string, value:NenvValue), nenv): NEnvironment =
            add_nbinding(nenv, name, value)
    in Assoc_List.iterate(combine_one, nenv1, nenv2) end

fun combine_tenvs(tenv1:tEnvironment, tenv2:tEnvironment):tEnvironment = 
    let fun combine_one((name:string, value:Type), tenv): tEnvironment =
            add_tbinding(tenv, name, value)
    in Assoc_List.iterate(combine_one, tenv1, tenv2) end

(* 
 *  NotFound is the exception raised if lookups or removes fail.
 *)
val NotFound = Assoc_List.NotFound

(* 
 *  lookup_binding takes an environment and a name, and returns the
 *  value bound to that name in the environment, or raises NotFound if the
 *  name was unbound.
 *)
fun lookup_binding(env:Environment, name:string):Value =
    Assoc_List.fetch(env, name)

fun lookup_nbinding(nenv:NEnvironment, name:string):NenvValue = 
    Assoc_List.fetch(nenv, name)

fun lookup_tbinding(tenv:tEnvironment, name:string):Type = 
    Assoc_List.fetch(tenv, name)

(* 
 *  printEnv prints out an environment.
 *)
fun printEnv(env:Environment):string =
    let 
        fun print_one((name:string, value:Value), rest_str):string =
	    "val " ^ name ^ " = " ^ printValue(value) ^ "\n" ^ rest_str
    in 
        Assoc_List.iterate(print_one, "", env) 
    end

fun printNEnv(nenv:NEnvironment):string =
    let 
        fun print_one((name:string, value:NenvValue), rest_str):string =
	    "val " ^ name ^ " = " ^ printNenvValue(value) ^ "\n" ^ rest_str
    in 
        Assoc_List.iterate(print_one, "", nenv) 
    end

fun printtEnv(tenv:tEnvironment):string = 
    let 
        fun print_one((name:string, value:Type), rest_str):string =
            "name " ^ name ^ " = " ^ printType(value) ^ "\n" ^ rest_str
    in
        Assoc_List.iterate(print_one, "", tenv)
    end
