(*
 *  eval.sml: This file is the third step in the ICE process.  When completed, it will
 *  recieve an internalized and type-checked decl or expr and evaluate it.
 *)

(* 
 *  if pattern-matching fails, it raises the MatchFailure exception 
 *)
exception MatchFailure

(* 
 *  eval() takes the top-level decl or expr, and the top-level environment
 *  (i.e., the bindings of global variables to their values), and then evaluates
 *  the top-level AST to produce a set of new bindings, which is returned to the
 *  caller
 *)
fun eval(ast:TopLevel, env:Environment):Environment =
    let
        val infoText = "[Evaluating " ^ printTopLevel(ast) ^ "]\n"
        val _ = emlPrint(infoText)
    in
        evalTopLevel(ast, env)
    end

(* 
 *  evalTopLevel() does the work of eval
 *)
and evalTopLevel(TopLevelValDeclIntern(decl), env):Environment =
        evalVDecl(decl, env)
  | evalTopLevel(TopLevelOODeclIntern(decl), env) =
        evalOODeclList(decl, env)
  | evalTopLevel(_) = 
(*
 *  Everything we see MUST be internalized by this time.  We should never
 *  encounter a non-internalized top-level at this point.
 *)
    raise EvalError("Evaluation has been passed a non-internalized " ^
        "top-level expression!  This should have been internalized in the " ^
        "internalization phase!  [This error should never occur.]")

(* 
 *  To evaluate a variable declaration, first evaluate its rhs expression
 *  to produce a result Value, then match the rhs value against the lhs pattern,
 *  returning any variables that are bound by the pattern-match.
 *)
and evalVDecl(ValDeclIntern{lhs, lhs_type, rhs}, env):Environment =
    let 
        val value:Value = evalExpr(rhs, env)
        val bindings:Environment = matchPattern(lhs, value)
	    handle MatchFailure =>
		raise EvalError("Pattern-match in val binding failed")
    in
        bindings
    end

  | evalVDecl(UnTypedValDeclIntern{lhs, rhs}, env) = 
(*
 *  At this point (the evaluation phase) there is no difference between a typed
 *  and an untyped variable declaration.  The code here is the same as for the
 *  previous case.
 *)
    let
        val value:Value = evalExpr(rhs, env)
        val bindings:Environment = matchPattern(lhs, value) 
            handle MatchFailure => 
                raise EvalError("Pattern-match in val binding failed")
    in
        bindings
    end

and evalOODecl(FunDeclIntern({name=internalName, iType=evalType, fcases}), env) =
(*
 *  To evaluate a function declaration, simply construct a value representing
 *  the function (along with the current environment, to support lexical scoping
 *  when the function is later called) and then return a binding of the function
 *  name to the function value.
 *)
    let
	val fun_value = FunValue{name=internalName, cases=fcases, env=env,
                dType=evalType}
    in
        new_binding(internalName, fun_value)
    end
  
  | evalOODecl(ClassDeclIntern(_), env) = 
(*
 *  All of the work is done in the internalize phase.  We don't keep track of
 *  the classes in the value environment.
 *)
    empty_environment    

  | evalOODecl(InterfaceDeclIntern(_), env) =
(*
 *  Evaluating an interface is like evaluating a class, since they are the same
 *  thing.
 *)
    empty_environment

  | evalOODecl(MessageDeclIntern({name=message_name, iType=evalType}), env) = 
(*  
 *  To create a new message simply create an empty method list and add a binding
 *  with the message name and the list.
 *)
    let 
        val emptyMethodList:Method list = nil 
	val message_value = GenericFunValue({mdList=ref emptyMethodList,
                dType=evalType})
    in
        new_binding(message_name, message_value)
    end

  | evalOODecl(MethodDeclIntern({name=msgName, formal=msgFormal, body=msgBody}),
        env) = 
(*
 *   To evaluate a method, we just take apart the method ref and then add our 
 *   method to that list.  We do not add any new bindings to the environment.
 *)
    let 
        fun getMethodListFromString(msg_name) = 
            let 
                val gfVal = (lookup_binding(env, msg_name)
		    handle NotFound => raise EvalError("Method name: " ^
                        msg_name ^ " not found in environment"))
                
                fun getGFFromValue(cl) = 
		    case cl of
		        GenericFunValue({mdList=methodList, dType=dt}) =>
                            methodList
		      | _ => raise EvalError("Method name: " ^ msg_name ^
                            " is not a valid Generic Function")
            in 
                getGFFromValue (gfVal) 
            end

        val prevMethodListRef = getMethodListFromString(msgName)
        val prevMethodList:Method list = !prevMethodListRef
	val newMethod:Method = {formal=msgFormal, body=msgBody, env=env}
	val newMethodList:Method list = newMethod :: prevMethodList
    in
	(prevMethodListRef := newMethodList;empty_environment)
    end
    
(*
 *  evalVDeclList() evaluates a list of declarations to produce a set of new
 *  bindings from all the declarations.
 *)
and evalVDeclList([], env) = empty_environment
  | evalVDeclList(first::rest, env) =
(*
 *  We evaluate the first declaration.  Then we evaluate the remaining
 *  declarations, but in the environment including the bindings from the first
 *  declaration.  We return the set of new bindings from the declaration list.
 *)
    let 
        val first_bindings:Environment = evalVDecl(first, env) 
        val extended_env:Environment = combine_envs(env, first_bindings)
        val rest_bindings:Environment = evalVDeclList(rest, extended_env)
    in 
        combine_envs(first_bindings, rest_bindings) 
    end

(*
 *  evalVDeclList() evaluates a list of declarations to produce a set of new
 *  bindings from all the declarations.
 *)
and evalOODeclList([], env) = empty_environment
  | evalOODeclList(first::rest, env) =
(*
 *  We evaluate the first declaration.  Then we evaluate the remaining
 *  declarations, but in the environment including the bindings from the first
 *  declaration.  We return the set of new bindings from the declaration list.
 *)
    let 
        val first_bindings:Environment = evalOODecl(first, env) 
        val extended_env:Environment = combine_envs(env, first_bindings)
        val rest_bindings:Environment = evalOODeclList(rest, extended_env)
    in 
        combine_envs(first_bindings, rest_bindings) 
    end

(*
 *  evalExpr() evaluates an expression in an environment to produce a result
 *  Value.
 *)
and evalExpr(VarExprIntern(var_name), env) =
(*
 *  To evaluate a variable reference, just look up the variable's binding in the
 *  current environment (raise an error if the variable isn't bound in the
 *  current environment).
 *)
    (lookup_binding(env, var_name) 
        handle NotFound => 
            raise EvalError("Variable " ^ var_name ^ " not bound."))

(*
 *  To evaluate a constant, just return the corresponding value.
 *)
  | evalExpr(IntExprIntern(v), env) = IntValue(v)
  | evalExpr(RealExprIntern(v), env) = RealValue(v)
  | evalExpr(StringExprIntern(v), env) = StringValue(v)
  | evalExpr(BoolExprIntern(v), env) = BoolValue(v)
  | evalExpr(NilExprIntern(decType), env) = ListValue(nil)
  | evalExpr(TupleExprIntern(exprs:ExprIntern list), env) =
(*
 *  To evaluate a tuple expression, evaluate each of the subexpressions, then
 *  construct the tuple value from those evaluations.
 *)
    TupleValue(evalExprList(exprs, env))

  | evalExpr(RecordExprIntern(recs), env) = 
(*
 *  To evaluate a record, we need to evaluate the expression in each element.
 *  We do this by looping through the list of elements, evaluated the 
 *  expressions as we go.
 *)
    let
        fun evalRecs(nil) = nil
          | evalRecs({eident=eID, evalue=eVal}::restRecs) = 
                {ident=eID, value=evalExpr(eVal, env)}::evalRecs(restRecs)
    in
        RecordValue(evalRecs(recs))
    end

  | evalExpr(UnopExprIntern{operator, arg}, env) =
(*
 *  To evaluate a unary operator expression, evaluate the argument, then apply
 *  the operator to the argument.
 *)
    let
        val arg_value = evalExpr(arg, env)
    in
        applyUnop(operator, arg_value)
    end

  | evalExpr(BinopExprIntern{arg1, operator, arg2}, env) =
(*
 *  To evaluate a binary operator expression, evaluate the arguments, then apply
 *  the operator to the arguments.
 *)
    let
        val arg1_value = evalExpr(arg1, env)
        val arg2_value:Value = evalExpr(arg2, env)
    in
        applyBinop(arg1_value, operator, arg2_value)
    end

  | evalExpr(CallExprIntern{function:ExprIntern, arg:ExprIntern}, env):Value =
(*
 *  To evaluate a function call, first find the function value in the
 *  environment (like a variable reference expression), and verify that its
 *  actually a function value.  Then evaluate the argument expression.  Finally,
 *  evaluate the function's cases (described below), in the environment
 *  lexically enclosing the callee function's declaration.
 *)
    let
        val fun_value = evalExpr(function, env)
        val arg_value = evalExpr(arg, env)
    in
        case fun_value of
	    FunValue({name=fn_name, cases=cases, env=fn_env, dType}) => 
            let
                val fn_env_allowing_recursion =
                        add_binding(fn_env, fn_name, fun_value)
	    in
(*
 *              fn_env is the environment lexically enclosing the function
 *              declaration, which is the one we want to evaluate the function
 *              body in.  using the current environment of the call, env, would
 *              be wrong, as it would allow the callee function to access the
 *              local variables visible at the call site, and it wouldn't allow
 *              the callee function to access any local variables visible where
 *              it was defined but not at the call site.  However, fn_env is the
 *              environment just *before* the function was declared, and so it
 *              doesn't include a binding for the function itself.  In case the
 *              function is recursive, we need to extend fn_env with a binding
 *              for the function, so that it can call itself.  That's what we do
 *              now.
 *)
                evalCaseList(cases, arg_value, fn_env_allowing_recursion) 
	    end

	  | ClassValue(parts) => 
(*
 *          To evaluate a class value, we create a class object and give it the
 *          correct parts.
 *)
            ClassObject{class=parts, rep=arg_value}

	  | GenericFunValue({mdList=msgList, dType=dt}) => 
(*
 *          To evalaute a generic function call, we must pick which method best
 *          fits the call.
 *)
            evalMethodList(fun_value, arg_value, env, !msgList)

	  | _  => raise EvalError("Encountered an invalid calling Value.")
    end
		  
  | evalExpr(IfExprIntern{testExpr, thenExpr, elseExpr}, env) =
(*
 *  To evaluate an if, first evaluate the testExpr.  If its true, evaluate and
 *  return the thenExpr, otherwise evaluate and return the elseExpr.
 *)
    let
        val test_value = evalExpr(testExpr, env)
        fun eval_if(BoolValue(true),thenExpr,_,env) = evalExpr(thenExpr, env)
          | eval_if(BoolValue(false),_,elseExpr,env) = evalExpr(elseExpr, env)
          | eval_if(_,_,_,env) = raise EvalError("Test expression isn't a bool")
    in
        eval_if(test_value, thenExpr, elseExpr, env)
    end

  | evalExpr(LetExprIntern{decls:vDeclIntern list, body}, env) =
(*
 *  To evaluate a let, first evaluate the decls to produce a set of new
 *  bindings.  Then extend the current environment "env" with the new bindings
 *  to produce a new extended environment.  Finally, evaluate the body
 *  expression in the environment including all the new bindings, and return the
 *  result of the body evaluation.
 *)
    let
        val bindings = evalVDeclList(decls, env)
        val new_env = combine_envs(env, bindings)
    in
        evalExpr(body, new_env)
    end

  | evalExpr(ClassExprIntern(evaledClass as Class{name, id, mode, superclass, 
        interfaces, dType}), env) = ClassValue(evaledClass)

(*
 *  newBestMatchList() takes a list and a method and determines which methods
 *  should still be in the list.
 *
 *  The first parameter is the new method that we are checking to see if it
 *      belongs in the list.
 *  The second parameter is the list that we are comparing the method against.
 *  The third parameter is the new list that we are creating.
 *)

and newBestMatchList(newMethod, nil, newList) =
(*
 *  If the original list is empty, then we can just add the last method on
 *  to the new list and return that.
 *)
    newMethod::newList

  | newBestMatchList(newMethod as {formal=f1,body=b1,env=e1}, (x as 
        {formal=f2,body=b2,env=e2})::xs, newList) = 
(*
 *  Otherwise we compare the specificity of the first element on the original
 *  list and the new method, and make a decision based on the results.
 *)
    case compareSpecificity(f1, f2) of 
        FirstIsMoreSpecific =>
(*
 *      If the first is more specific, then we can drop the element from the
 *      list and compare the new method against the rest of the list.
 *)
        newBestMatchList(newMethod, xs, newList)

      | SecondIsMoreSpecific =>
(*
 *      If the second is more specific, then we can stop looking and put the
 *      list back together because the new method is not as specific as the
 *      methods in the list.
 *)
        x::xs @ newList

      | _ =>
(*
 *      In any other case (they have the same specificity or they are
 *      incomperable), keep the method in the list and check the rest of the
 *      list.
 *)
        newBestMatchList(newMethod, xs, x::newList)

(*
 *  evalMethodList() determines which method best fits a generic function call.
 *)
and evalMethodList(fun_val, arg_val, env, methodList) = 
    let
        fun evalBestMatch((method as {formal,body,env})::methods, curBestList) =
(*
 *          If the method pattern matches with the value, then determine which
 *          values should remain in the list.  If it doesn't match, just use the
 *          current list (without changing).
 *)
            let
                val newMethodList = if matchPatternBool(formal, arg_val) then
                        newBestMatchList(method, curBestList, nil)
                    else curBestList
            in
                evalBestMatch(methods, newMethodList)
            end

          | evalBestMatch(nil, curBestList) = 
(*
 *          If our methodList is empty, then check to see if we have the 1 and
 *          only 1 method in our best fit list.
 *)
            case length(curBestList) of 
                0 => raise EvalError("No matching cases found")
              | 1 => 
                    let
                        val meth as {formal=gf_f,body=gf_b,env=gf_e} =
                            hd(curBestList)
                        val _ = print("DEBUG: Finished getting best case\n")
                    in
                        evalCaseList([{formal=gf_f,body=gf_b}], arg_val, gf_e)
                    end
              | _ => raise EvalError("More than 1 best case found")
    in
        evalBestMatch(methodList,nil)
    end

(*
 *  evalExprList() evaluates a list of expressions to produce a list of Values.
 *)
and evalExprList([]:ExprIntern list, env):Value list = []
  | evalExprList(first::rest, env):Value list =
        evalExpr(first, env)::evalExprList(rest, env)

(* 
 *  evalCaseList() loops over a list of function cases, in order, to find the 
 *  first one whose formal argument pattern matches the argument value.  When 
 *  one matches, yielding a list of bindings from the match, the body of that 
 *  case is evaluated in the *global* environment (since we don't want the 
 *  callee function to be able to reference local variables of the calling 
 *  function) extended with the formal's bindings. 
 *)
and evalCaseList([], arg:Value, env):Value =
(* 
 *  If we run out of cases, then we've hit a match-not-exhaustive error.
 *)
    raise EvalError("No matching case for function call")

  | evalCaseList({formal, body}::rest, arg:Value, env) =
(* 
 *  Try the first case, and if it successfully pattern-matches (i.e. doesn't
 *  raise a MatchFailure exception) then proceed to evaluate that case's 
 *  body expression.
 *)
    let
        val formal_bindings:Environment = matchPattern(formal, arg)
	val fun_env:Environment = combine_envs(env, formal_bindings)
    in
        evalExpr(body, fun_env)
    end
(* 
 *  If the pattern-match on the first case's formal failed, then proceed 
 *  to the next case.
 *)
    handle MatchFailure => evalCaseList(rest, arg, env)

(* 
 *  matchPattern() takes a pattern and a value and does pattern-matching.  If
 *  successful, it returns a list of bindings of variables in the pattern to
 *  their corresponding pieces of the matched value.  If unsuccessful, it raises
 *  a MatchFailure exception.
 *)
and matchPattern(VarPatternIntern(name:string), value):Environment =
(*
 *  Variable patterns always succeed, and make a binding for the variable to
 *  the value that it matched.
 *)
    new_binding(name, value)

  | matchPattern(WildcardPatternIntern, value):Environment =
(* 
 *  Wildcard patterns always succeed, but don't create any bindings.
 *)
    empty_environment

(*
 *  Constant patterns match equal constant values, without creating any 
 *  bindings.
 *)
  | matchPattern(IntPatternIntern(i1), IntValue(i2)):Environment =
        if i1 = i2 then empty_environment else raise MatchFailure
  | matchPattern(RealPatternIntern(r1), RealValue(r2)) =
        if r1 <= r2 andalso r1 >= r2 then empty_environment else raise MatchFailure
  | matchPattern(StringPatternIntern(s1), StringValue(s2)):Environment =
        if s1 = s2 then empty_environment else raise MatchFailure
  | matchPattern(BoolPatternIntern(b1), BoolValue(b2)):Environment =
        if b1 = b2 then empty_environment else raise MatchFailure
  | matchPattern(NilPatternIntern, ListValue(nil)):Environment =
        empty_environment

  | matchPattern(RecordPatternIntern(patRecList), RecordValue(valRecList)) =
(*
 *  To match against a record, we must make sure that each member in the pattern
 *  has a matching value.
 *)
    let
(*
 *      matchRecPats() takes a list of pattern records and a list of value
 *      records and compares the two lists, returning the env created by the
 *      pattern matching at each element.
 *)
        fun matchRecPats(nil, nil) = empty_environment
          | matchRecPats(nil, notNil) = raise EvalError("Record pattern has "
                ^ "fewer elements than the record's value.")
          | matchRecPats(notNil, nil) = raise EvalError("Record pattern has "
                ^ "more elements than the record's value.")
          | matchRecPats({ident=pId, pattern=pPat}::pRest,
                {ident=vId, value=vVal}::vRest) =
(*
 *          To match one element, we first make sure that the two labels are
 *          the same.  We then pattern match the value and combine the env
 *          created with the env of the rest of the patterns.
 *)
            if (pId = vId) then
            let
                val new_env = matchPattern(pPat, vVal)
            in
                combine_envs(new_env, matchRecPats(pRest, vRest))
            end
            else raise EvalError("Record pattern name: " ^ pId ^ 
                " doesn't match value name: " ^ vId ^ ".")
    in
        matchRecPats(patRecList, valRecList)
    end

  | matchPattern(ClassPatternIntern({class=patClass, rep=patRep}), 
                 ClassObject({class=valClass, rep=valRep})) =
(*
 *  To pattern match two classes, we first make sure that the value class is a
 *  subclass of the pattern class.  If so, we can pattern match on the two
 *  representations to make sure that they match and return the results of that
 *  pattern match.
 *)
    if (FirstIsSubclassOfSecond(valClass, patClass)) then
        matchClassPattern(patRep, valRep)
    else raise MatchFailure

  | matchPattern(TuplePatternIntern(patterns), TupleValue(values)) =
(*
 *  A tuple pattern matches a tuple value if its components match.
 *)
    matchPatterns(patterns, values)

  | matchPattern(ConsPatternIntern{head, tail}, ListValue(first::rest)) =
(*
 *  A :: pattern matches a non-empty list value if its head and tail patterns
 *  match the first and rest elements of the list, respectively.
 *)
    let
        val head_bindings = matchPattern(head, first)
	val tail_bindings = matchPattern(tail, ListValue(rest))
    in
        combine_envs(head_bindings, tail_bindings)
    end
    
  | matchPattern(_, _) =
(*
 *  Otherwise, things don't match.
 *)
    raise MatchFailure

(*
 *  If the pattern match succeeds, then return true.  If match fails (via an
 *  exception), then return false.
 *)
and matchPatternBool(pattern, arg_val) =
    (matchPattern(pattern, arg_val);true) handle MatchFailure => false

(*
 *  matchPatterns() matches a list of patterns against a list of values,
 *  returning a list of bindings resulting from all the matches.
 *)
and matchPatterns([]:PatternIntern list, []:Value list):Environment =
(*
 *  empty lists match, without producing any bindings.
 *)
    empty_environment

  | matchPatterns(first_pattern::rest_patterns, first_value::rest_values) =
(*
 *  Matching a non-empty list of patterns against a non-empty list of values
 *  first proceeds recursively, combining the bindings derived from the matches
 *  of two subparts.
 *)
    let
        val first_bindings = matchPattern(first_pattern, first_value)
        val rest_bindings = matchPatterns(rest_patterns, rest_values)
    in
        combine_envs(first_bindings, rest_bindings)
    end

  | matchPatterns(_, _) =
(* 
 *  Lists of patterns and values that aren't of equal length don't match.
 *)
    raise MatchFailure

(*
 *  matchClassPattern() is the same as matchPattern() except that it allows the 
 *  first list of patterns to be shorter (for subtypes).  We know that the pat
 *  and the value must be Records to be a valid class type (other cases an error
 *  is raised.
 *)
and matchClassPattern(RecordPatternIntern(recPat), RecordValue(recVal)) = 
(*
 *      matchOneElt() takes a record pattern element and the record's value list
 *      and finds the corresponding element in that list.  When the element is
 *      found, it pattern matches the pattern of the element with the found type.
 *      If no match is found, and error is raised.
 *)
    let
        fun matchOneElt(elt as {ident=id, pattern=pat}, nil) = 
(*
 *          The list of types is empty, and no match was found for the name, so
 *          an error is raised.
 *)   
            raise EvalError("Found no match for: " ^ id ^ " in record's value.")

          | matchOneElt(elt as {ident=pID, pattern=pPat},
                {ident=eID, value=eValue}::rest) =
            if (pID = eID) then
(*
 *              The names match, so we type check the pattern, return the env
 *              created from the type check.
 *)
                matchPattern(pPat, eValue)
            else
(*
 *              The names don't match, so check the rest of the list.
 *)
                matchOneElt(elt, rest)

(*
 *      matchAllElts() takes the list of the record's pattern elements and
 *      checks each one against the value list, using matchOneElt(), and returns
 *      the catenation of all the type environments created from the
 *      matchOneElt() calls.
 *)
        fun matchAllElts(nil, valList) = empty_environment
          | matchAllElts(first::rest, valList) = 
                combine_envs(matchOneElt(first, valList),
                    matchAllElts(rest, valList))
    in
        matchAllElts(recPat, recVal)
    end

(*
 *  We can also handle a wildcard case.
 *)
  | matchClassPattern(WildcardPatternIntern, RecordValue(recVal)) =
        matchPattern(WildcardPatternIntern, RecordValue(recVal))

(*
 *  We can also handle a variable pattern case.
 *)
  | matchClassPattern(VarPatternIntern(vname), RecordValue(recVal)) =
        matchPattern(VarPatternIntern(vname), RecordValue(recVal))

(*
 *  If we don't have records, then we raise an error.
 *)  
  | matchClassPattern(RecordPatternIntern(_), _) =
     raise EvalError("matchClassPattern() errornously encountered a " ^
        "non-RecordValue.")
  | matchClassPattern(_, RecordValue(_)) =
     raise EvalError("matchClassPattern() errornously encountered a " ^
        "non-RecordPattern.")
  | matchClassPattern(_, _) =
     raise EvalError("matchClassPattern() errornously encountered a " ^
        "non-RecordPattern and non-RecordValue.")

(*
 *  applyUnop() and applyBinop() carry out the work of applying an operator to
 *  its argument value(s), returning a result value.  They raise EvalError if
 *  the type(s) of the argument value(s) are not appropriate for the operator.
 *
 *  <colin> 12/10/2001 -- We want to remove this code some day and have these
 *  functions *not* hard-coded into EML.
 *)
and applyUnop("~", IntValue(arg)) = IntValue(~arg)
  | applyUnop("~", RealValue(arg)) = RealValue(~arg)
  | applyUnop("not", BoolValue(arg)):Value = BoolValue(not arg)
  | applyUnop(unop, v) = raise EvalError("Value " ^ printValue(v) ^ " mistyped "
        ^ " argument to operator " ^ unop ^ ".")

and applyBinop(IntValue(arg1), "+", IntValue(arg2)) = IntValue(arg1 + arg2)
  | applyBinop(RealValue(arg1), "+", RealValue(arg2)) = RealValue(arg1 + arg2)
  | applyBinop(IntValue(arg1), "-", IntValue(arg2)) = IntValue(arg1 - arg2)
  | applyBinop(RealValue(arg1), "-", RealValue(arg2)) = RealValue(arg1 - arg2)
  | applyBinop(IntValue(arg1), "*", IntValue(arg2)) = IntValue(arg1 * arg2)
  | applyBinop(RealValue(arg1), "*", RealValue(arg2)) = RealValue(arg1 * arg2)
  | applyBinop(IntValue(arg1), "div", IntValue(arg2)) =
        IntValue(arg1 div arg2)
  | applyBinop(RealValue(arg1), "/", RealValue(arg2)) = RealValue(arg1 / arg2)
  | applyBinop(IntValue(arg1), "mod", IntValue(arg2)) =
        IntValue(arg1 mod arg2)
  | applyBinop(IntValue(arg1), "<",  IntValue(arg2)) = BoolValue(arg1 <  arg2)
  | applyBinop(RealValue(arg1), "<",  RealValue(arg2)) = BoolValue(arg1 <  arg2)
  | applyBinop(StringValue(arg1), "<",  StringValue(arg2)) = BoolValue(arg1 <  arg2)
  | applyBinop(IntValue(arg1), "<=", IntValue(arg2)) = BoolValue(arg1 <= arg2)
  | applyBinop(RealValue(arg1), "<=", RealValue(arg2)) = BoolValue(arg1 <= arg2)
  | applyBinop(StringValue(arg1), "<=", StringValue(arg2)) = BoolValue(arg1 <= arg2)
  | applyBinop(IntValue(arg1), ">=", IntValue(arg2)) = BoolValue(arg1 >= arg2)
  | applyBinop(RealValue(arg1), ">=", RealValue(arg2)) = BoolValue(arg1 >= arg2)
  | applyBinop(StringValue(arg1), ">=", StringValue(arg2)) = BoolValue(arg1 >= arg2)
  | applyBinop(IntValue(arg1), ">",  IntValue(arg2)) = BoolValue(arg1 >  arg2)
  | applyBinop(RealValue(arg1), ">",  RealValue(arg2)) = BoolValue(arg1 >  arg2)
  | applyBinop(StringValue(arg1), ">",  StringValue(arg2)) = BoolValue(arg1 >  arg2)
  | applyBinop(StringValue(arg1), "^", StringValue(arg2)) =
        StringValue(arg1 ^ arg2)
  | applyBinop(arg1, "::", ListValue(arg2)) = ListValue(arg1 :: arg2)
  | applyBinop(v1, "=", v2) = BoolValue(is_equal(v1, v2))
  | applyBinop(v1, "<>", v2) = BoolValue(not(is_equal(v1, v2)))
  | applyBinop(v1, binop, v2) = raise EvalError("Values (" ^ printValue(v1) ^
        "; " ^ printValue(v2) ^ ") mistyped arguments to " ^ binop ^ ".")


(*
 *  is_equal is a helper function that implements the = operator on pairs of
 *  values of the same type.  it raises an EvalError if the arguments are of
 *  different types of if functions are compared.
 *)
and is_equal(IntValue(arg1), IntValue(arg2)) = arg1 = arg2
  | is_equal(StringValue(arg1), StringValue(arg2)) = arg1 = arg2
  | is_equal(BoolValue(arg1), BoolValue(arg2)) = arg1 = arg2
  | is_equal(TupleValue(values1), TupleValue(values2)) =
        are_equal(values1, values2)
  | is_equal(ListValue(values1), ListValue(values2)) =
        are_equal(values1, values2)
  | is_equal(FunValue(arg1), FunValue(arg2)) = 
        raise EvalError("cannot apply = or <> to a function value")
  | is_equal(v1, v2) = raise EvalError("Values (" ^ printValue(v1) ^ "; " ^
        printValue(v2) ^ ") mistyped arguments to = or <>")

and are_equal([], []) = true
  | are_equal(first1::rest1, first2::rest2) =
        is_equal(first1, first2) andalso are_equal(rest1, rest2)
  | are_equal(_, _) = false
