(*
 *  ICE: Internalize, Check, and Evaluate.  This is performed on everything that
 *  is entered into the MiniEML parser.
 *
 *  iceControl.sml is the top level that calls each step in the right order.
 *)
fun ice(ast:TopLevel, nenv, tenv, itcenv, env, EMLenv) = 
    let

(*
 *      doICE() executes the internalize, type check and evaluation.  This is
 *      called after we made sure that the TopLevel was not a call that EML
 *      needed to handle.
 *
 *      doICE() returns new environments created by internalizing, type
 *      checking, and evaluation, as well as the original EML environment
 *      (since the EML env hasn't changed).
 *)
        fun doICE() = 
        let
            val (astInternalized, new_nenv) = internalize(ast, nenv)
            val (new_tenv, new_itcenv) = typeCheck(astInternalized, tenv, itcenv,
                combine_nenvs(nenv, new_nenv))
            val new_env = eval(astInternalized, env)
        in
            (new_nenv, new_tenv, new_itcenv, new_env, EMLenv)
        end

(*
 *      checkEMLInterp() checks the new TopLevel (ast) to see if it is a
 *      toplevel that the interpretter needs to handle (a TopLevelSysCall).
 *      If it is, it handles it and returns true, signalling that the TopLevel
 *      has been handled and it doesn't need to go through the ICE process.
 *      If returning true, it also passes back a new EML input stream
 *      environment.  This environment will replace the old one only if this
 *      returns true.
 *)
        fun checkEMLInterp(curEMLenv) =
            case (ast) of
                TopLevelSysCall(SysCallVerbose(vSetting)) => 
                let
                    val _ = emlSetVerbose(vSetting)
                in
                    (true, curEMLenv)
                end
              | TopLevelSysCall(SysCallUse(useFile)) =>
                let
                    val useFp = TextIO.openIn(useFile)
                        handle Io => raise EvalError("Could not open file: " ^
                        useFile)
                    val newEMLenv = {instream=useFp, fromFile=1} :: curEMLenv
                in
                    (true, newEMLenv)
                end
              | _ => (false, nil)
        
(*
 *      checkICE() checks whether the TopLevel is an EML specific call (using
 *      checkEMLInterp()).  If it is, it returns empty environments the new
 *      EML environment that might have been created by the call.  Otherwise
 *      just call the ICE process (using doICE), which returns the new envs
 *      created by the process.
 *)
        fun checkICE() = 
        let
            val (isEMLInterpLine, new_EMLenv) = checkEMLInterp(EMLenv)
        in
            if (isEMLInterpLine) then (empty_nenvironment,
                empty_tenvironment, itcenv, empty_environment, new_EMLenv)
            else doICE()
        end

    in
        checkICE()
    end
