(*
 *  emlInternalize.sml:  Converts code from the declared form to our internal 
 *  format.  The conversion is mostly changing class names to their internal 
 *  representation of a Class structure.
 *)

fun internalize(ast:TopLevel, nenv:NEnvironment):(TopLevel * NEnvironment) =
    let
        val infoText = "[Converting declared " ^ printTopLevel(ast) ^ " to " ^
                       "internal format.]\n"
        val _ = emlPrint(infoText)
    in
        internTopLevel(ast, nenv)
    end

(* 
 *  internTopLevel does the work of internalize.  It takes either a val or oo
 *  declaration and reterns an internalized version.
 *
 *  A declaration results in both an internalized declaration and also a 
 *  name environment created from that declaration. 
 *)
and internTopLevel(TopLevelValDecl(vdecl), nenv) =
    let 
        val (new_decl, new_nenv) = internVDecl(vdecl, nenv)
    in
        (TopLevelValDeclIntern(new_decl), new_nenv)
    end
  | internTopLevel(TopLevelOODecl(oodecl), nenv) =
    let 
        val (new_decl, new_nenv) = internOODeclList(oodecl, nenv)
    in
        (TopLevelOODeclIntern(new_decl), new_nenv)
    end
  | internTopLevel(_, _) = raise InternError("Internalization recieved a top "
        ^ "level declaration that was not a val nor OO decl.")

(*
 *  To internalize a ValDecl, we must internalize the pattern, the type, and the
 *  rhs expression.  The new name environment is the name environment that is 
 *  returned from the internalized pattern.
 *)
and internVDecl(ValDecl({lhs=dec_lhs:Pattern,lhs_type=l_type,rhs=dec_rhs:Expr}),
    nenv) = 
    let
        val (new_lhs, pat_nenv) = internPat(dec_lhs, nenv)
        val intern_lhs_type = internType(l_type, nenv)
(*
 *      We combine the envs because the body might use variables declared
 *      in the pattern.
 *)
        val expanded_nenv = combine_nenvs(nenv, pat_nenv)
        val new_rhs = internExpr(dec_rhs, expanded_nenv)
    in
        (ValDeclIntern({lhs=new_lhs, lhs_type=intern_lhs_type, rhs=new_rhs}),
        pat_nenv)
    end

  | internVDecl(UnTypedValDecl({lhs=dec_lhs:Pattern, rhs=dec_rhs:Expr}), nenv) =
    let
        val (new_lhs, pat_nenv) = internPat(dec_lhs, nenv)
        val expanded_nenv = combine_nenvs(nenv, pat_nenv)
        val new_rhs = internExpr(dec_rhs, expanded_nenv)
    in
        (UnTypedValDeclIntern({lhs=new_lhs, rhs=new_rhs}), pat_nenv)
    end

(*
 *  To intern a function, we need to internalize the type and the cases.  The 
 *  name environment identifies the function name as not a class.
 *)
and internOODecl(FunDecl({name=dec_name,decType=dec_type,fcases=dec_cases}), 
    nenv) =
    let 
        val new_type = internType(dec_type, nenv)

        (*
         *  internCases() takes in a list of patterns and expressions, and 
         *  returns the list with the patterns and expression internalized.
         *)
        fun internCases(nil, nenv) = nil
          | internCases({body=exprBody, formal=fml} :: rest, nenv) =
                let
                    val (new_pat, pat_nenv) = internPat(fml, nenv)
                    val new_nenv = combine_nenvs(nenv, pat_nenv) 
                    val new_expr = internExpr(exprBody, new_nenv)
                in
                    {body=new_expr, formal=new_pat} :: internCases(rest, nenv)
                end
        val new_cases = internCases(dec_cases, nenv)
        val new_nenv = new_nbinding(dec_name, ntNotAClass)
    in
        (FunDeclIntern({name=dec_name,iType=new_type,fcases=new_cases}),
        new_nenv)
    end

(* 
 *  To internalize a Class, we must do the following:
 *   (1) Convert the superclass name to a superclass Class
 *   (2) Convert the interface name list to an interface Class list
 *   (3) Convert the type from unevaluated to evaluated.
 *  We then return a new declaration, ClassDeclIntern, which has the internalized
 *  information. 
 *)
  | internOODecl(ClassDecl({name=class_name,mode=class_mode,superclass=sclass_name,
    interfaces=interface_names,dType=unevalType}), nenv) =
    let 
        (* 
         *  (1) Convert the superclass name to a superclass Class
         *
         *  First take the superclass string option and retrieve the Class option
	 *  from this.  We must check to see if the name exists in the
         *  environment.  Once we have the Class structure we check to make sure
         *  that the superclass isn't an Interface.
         *)
        fun getSuperClassFromString(NONE) = NONE
	  | getSuperClassFromString(SOME cl_name) = 
	    let val sVal:NenvValue = (lookup_nbinding(nenv, cl_name)
		handle NotFound => raise InternError("Superclass " ^ cl_name ^ 
                                       " not found in name environment"))

		fun getClassFromValue(cl) = 
		    case cl of
                        ntClass(clOp as Class{name,id,mode=Interface,
                            superclass,interfaces,dType}) => raise 
                            InternError("Superclass " ^ name ^ " is an Interface")
                      | ntClass(clOp as Class(_)) => SOME(clOp)
                      | _ => raise InternError("Superclass is not a valid class")
	    in getClassFromValue (sVal) end

        (* 
         *  (2) Convert the interface name list to an interface Class list 
         *
         *  A similar function is used to get the interface class list from a 
         *  string list.  The functions are basically the same, only the 
         *  superclass function deals with a string option, while this one works
         *  with a string list. 
         *)
        fun getInterfaceClassesFromStringList(nil) = nil
          | getInterfaceClassesFromStringList(x::xs) =
	    let val iVal:NenvValue = (lookup_nbinding(nenv, x)
	        handle NotFound => raise InternError("Interface name " ^ x ^ 
                                       " not found in name envirnment."))
	        fun getClassFromValue(value) =
		    case value of
		        ntClass(clOp as Class({name,id,mode=Interface,
			    superclass,interfaces,dType})) => clOp
                      | _ => raise InternError("Declared interface name " ^ x ^
                                 " is not an Interface")
	    in 
	        getClassFromValue(iVal) :: getInterfaceClassesFromStringList(xs)
            end

        (* 
         *  sClass_class is the class option of the superclass
         *)
        val sClass_class:Class option = getSuperClassFromString(sclass_name)

        (* 
         *  iClass_classes is the list of interfaces
         *)
	val iClass_classes:Class list = 
            getInterfaceClassesFromStringList(interface_names)

        (* 
         *  (3) Convert the type from unevaluated to evaluated
         *
         *  First internalize the type, then combine it with the superclass' 
         *  type.
         *)
        val eType = internTypeOption(unevalType, nenv)

        fun combineTypes(SOME(localType)) =
            let 
                fun combineTypeHelper(lType) =
                case sClass_class of 
                    NONE => SOME(lType)
                  | SOME(Class({name,id,mode,superclass,interfaces,
                        dType=SOME(sType)})) => SOME(coalesceClassTypes(sType,lType))
                  | _ => raise InternError("Superclass for " ^ class_name ^ 
                             " has no type!")
            in 
                combineTypeHelper(localType)
            end
          | combineTypes(NONE) = 
            (* 
             *  If this class has no type, an error is raised.  Classes need to 
             *  have a type.
             *)
            raise InternError("Class " ^ class_name ^ " has no type!")

        val combinedInternType = combineTypes(eType)

        (*
         *  Now that we have all the info we need, we can create the Class 
         *  structure.
         *)
        val curId = getNextId()
        val cl_class = Class{name=class_name, id=curId, 
	    mode=class_mode, superclass=sClass_class,
            interfaces=iClass_classes, dType=combinedInternType}
        val cl_value:NenvValue = ntClass(cl_class)
        val new_nenv = new_nbinding(class_name, cl_value)
    in
        (ClassDeclIntern(cl_class), new_nenv)
    end

(* 
 *  Internalizing an interface is similar to evaluating a class (since they are 
 *  both represented by Class structures).  However here we only need to worry
 *  about checking the interface list, since there is no superclass nor type.
 *)
  | internOODecl(InterfaceDecl({name=interface_name, interfaces=interface_names}), 
    nenv) =
    let 
        fun getInterfaceClassesFromStringList(nil) = nil
	  | getInterfaceClassesFromStringList(x::xs) =
	    let val iVal:NenvValue = (lookup_nbinding(nenv, x)
                handle NotFound => raise InternError("Interface name " ^ x ^ 
                                       " not found in name env."))
		fun getClassFromValue(value) =
		    case value of
			ntClass(clOp as Class({name,id,mode=Interface,
			    superclass,interfaces,dType})) => clOp
		      | _ => raise InternError("Interface of " ^ x ^ " is not " ^
                                 "an Interface")
	    in 
		getClassFromValue(iVal) :: getInterfaceClassesFromStringList(xs)
	    end

	val iClass_classes:Class list = 
            getInterfaceClassesFromStringList(interface_names)

        val curId = getNextId()
        val interface_class = Class{name=interface_name, id=curId,
	    mode=Interface, superclass=NONE, interfaces=iClass_classes,
            dType=NONE}
        val interface_value:NenvValue = ntClass(interface_class)
        val new_nenv = new_nbinding(interface_name, interface_value)

    in
        (InterfaceDeclIntern(interface_class), new_nenv)
    end

(* 
 *  To internalize a message we need to internalize the type.
 *)
  | internOODecl(MessageDecl({name=message_name,dType=unevalType}), nenv) =
    let
        val new_type = internType(unevalType, nenv)
        val new_nenv = new_nbinding(message_name, ntNotAClass)
    in
        (MessageDeclIntern({name=message_name,iType=new_type}), new_nenv)
    end

(*
 *  To internalize a method, we need to make sure that there exists a message 
 *  with the same name.  We then internalize the pattern and the expression.
 *)
  | internOODecl(MethodDecl({name=message_name,formal=msgFormal,body=msgBody}),nenv) =
    let 
        val varNType  = (lookup_nbinding(nenv, message_name)
            handle NotFound => raise InternError("Message name " ^ message_name ^ 
                                   " not found in name environment"))
        
        fun validateName() = 
            if (varNType = ntNotAClass) then message_name
            else raise InternError("Message name " ^ message_name ^ 
                     " is a class!")

        val (new_pat, pat_nenv) = internPat(msgFormal, nenv)
        val expanded_nenv = combine_nenvs(nenv, pat_nenv)
        val new_expr = internExpr(msgBody, expanded_nenv)
    in
        (MethodDeclIntern({name=validateName(), formal=new_pat, body=new_expr}),
        empty_nenvironment)
    end

(*
 *  internVDeclList() takes a list of declarations and internalizes each of them.
 *)
and internVDeclList([], nenv) = (nil, empty_nenvironment)
  | internVDeclList(first::rest, nenv) = 
(*
 *  First, we internalize the first declaration.  Then we internalize the
 *  remaining declarations (recursively), but in the environment including the
 *  bindings from the first declaration.  We return the set of new bindings from
 *  the decl list.
 *)
    let 
        val (first_decl, first_nenv) = internVDecl(first, nenv)
        val (rest_decl, rest_nenv) = internVDeclList(rest, combine_nenvs(nenv,
            first_nenv))
    in
        (first_decl::rest_decl, combine_nenvs(first_nenv, rest_nenv))
    end

(*
 *  internOODeclList() takes a list of declarations and internalizes each of them.
 *)
and internOODeclList([], nenv) = (nil, empty_nenvironment)
  | internOODeclList(first::rest, nenv) = 
(*
 *  First, we internalize the first declaration.  Then we internalize the
 *  remaining declarations (recursively), but in the environment including the
 *  bindings from the first declaration.  We return the set of new bindings from
 *  the decl list.
 *)
    let 
        val (first_decl, first_nenv) = internOODecl(first, nenv)
        val (rest_decl, rest_nenv) = internOODeclList(rest, combine_nenvs(nenv,
            first_nenv))
    in
        (first_decl::rest_decl, combine_nenvs(first_nenv, rest_nenv))
    end


(*
 *  Internalize an expression.  For the basic types, we simply need to change the
 *  name from *Expr() to an *ExprInternalized().
 *)
and internExpr(VarExpr(var_name:string), nenv) = 
(*
 *  When internalizing a variable, we check to see if the variable corresponds
 *  to a class.  If it does, we replace the string with the class itself.
 *)
    let
        val varNType = (lookup_nbinding(nenv, var_name)
            handle NotFound => raise InternError("Variable " ^ var_name ^
                " not found in name environment"))
    in
        if (varNType = ntNotAClass) then
            VarExprIntern(var_name)
        else
        let
            val ntClass(cl_class) = varNType
        in
            ClassExprIntern(cl_class)
        end
    end

(*
 *  No work needed to internalize these simply experessions.
 *)
  | internExpr(IntExpr(v), nenv) = IntExprIntern(v)
  | internExpr(RealExpr(v), nenv) = RealExprIntern(v)
  | internExpr(StringExpr(v), nenv) = StringExprIntern(v)
  | internExpr(BoolExpr(v), nenv) = BoolExprIntern(v)
  | internExpr(NilExpr(decType), nenv) = 
(*
 *  We need to internalize the type of the expression and create an internalized
 *  nil expr with the new type.
 *)
    let 
        val new_type = internType(decType, nenv)
    in
        NilExprIntern(new_type)
    end
 
  | internExpr(TupleExpr(exprs), nenv) = 
    let
        val new_tuple = internExprList(exprs, nenv)
    in
        TupleExprIntern(new_tuple)
    end

  | internExpr(RecordExpr(recs), nenv) = 
    let
(*
 *  When we internalize a record, we need to internalize each element in the
 *  record.  We also alphabetize the elements while we do this.  Alphabetizing
 *  allows us to compare records more easily.
 *)

(*
 *      Inserts one element into an alphabetized list.
 *)   
        fun alphabetizeInsert(elt, eltsBefore, nil) = eltsBefore @ [elt]
          | alphabetizeInsert(elt as {eident=eltID, evalue=eltVal}, 
                eltsBefore, (eltCheck as {eident=checkID, evalue=checkval})
                ::eltsRest) =
            if (eltID < checkID) then eltsBefore @ [elt] @ (eltCheck::eltsRest)
            else alphabetizeInsert(elt, eltsBefore @ [eltCheck], eltsRest)

(*
 *      Creates an alphabetized list from the input list
 *)             
        fun alphabetizeList(nil, newList) =  newList
          | alphabetizeList(firstRec::restRecs, newList) =
            let
                val updList = alphabetizeInsert(firstRec, nil, newList)
            in
                alphabetizeList(restRecs, updList)
            end

(*
 *      Alphabetize list by the eident (string) element.
 *)    
        val alphaList = alphabetizeList(recs, nil)

(*
 *      Interns the expr in each element in the record list, returning a new list.
 *)
        fun internRecList(nil, newList) = newList
          | internRecList({eident=eID, evalue=eVal}::restRecs, newList) =
            let 
                val newRec = {eident=eID, evalue=internExpr(eVal, nenv)}
            in
                internRecList(restRecs, newList @ [newRec])
            end
    in
        RecordExprIntern(internRecList(alphaList, nil))
    end

  | internExpr(UnopExpr({operator=oper,arg=ar}), nenv) = 
(*
 *  We internalize the argument to the operator.
 *)
    let
        val new_expr = internExpr(ar, nenv)
    in
        UnopExprIntern({operator=oper,arg=new_expr})
    end

  | internExpr(BinopExpr({arg1=ar1,operator=oper,arg2=ar2}), nenv) = 
(*
 *  We internalize the arguments to the operator.
 *)
    let
        val new_expr1 = internExpr(ar1, nenv)
        val new_expr2 = internExpr(ar2, nenv)
    in
        BinopExprIntern({arg1=new_expr1,operator=oper,arg2=new_expr2})
    end

  | internExpr(CallExpr({function=func,arg=ar}), nenv) = 
(*
 *  At this point a CallExpr can be a class or a function, so we check to see
 *  if the CallExpr is an abstract class (if the CallExpr is a function, then
 *  we say it is not abstract).
 *)
    let
(*
 *      Returns true if the class passes the check (that is, the class is not
 *      abstract).  Also returns true if we are not passed a ClassExpr.
 *)
        fun abstractCallCheck(ClassExprIntern(Class{name, id, mode=Abstract, 
            superclass, interfaces, dType})) = false
          | abstractCallCheck(_) = true

(*
 *      Internalize the two parts of a CallExpr.
 *)
        val fun_expr = internExpr(func, nenv)
        val arg_expr = internExpr(ar, nenv)
    in
        if (abstractCallCheck(fun_expr)) then 
            CallExprIntern({function=fun_expr, arg=arg_expr})
        else raise InternError("Cannot create instance of abstract class")
    end

  | internExpr(IfExpr({testExpr=tstE,thenExpr=thenE,elseExpr=elseE}), nenv) =
(*
 *  Just internalize the three parts of the IfExpr.
 *)   
    let 
        val test_expr = internExpr(tstE, nenv)
        val then_expr = internExpr(thenE, nenv)
        val else_expr = internExpr(elseE, nenv)
    in
        IfExprIntern({testExpr=test_expr,thenExpr=then_expr,elseExpr=else_expr})
    end

  | internExpr(LetExpr({decls=ds,body=bds}), nenv) = 
(*
 *  First internalize all the declarations and then internalize the expression
 *  using a combined environment (which includes the new declarations' env).
 *)
    let
        val (let_decs, let_nenv) = internVDeclList(ds, nenv)
        val combined_nenv = combine_nenvs(nenv, let_nenv)
        val body_expr = internExpr(bds, combined_nenv)
    in
        LetExprIntern({decls=let_decs,body=body_expr})
    end

(*
 *  internalizeExprList() evaluates a list of expressions to produce a list of
 *  internalized exprs.
 *)
and internExprList([]:Expr list, nenv) = []
  | internExprList(first::rest, nenv) =
    let 
        val first_expr = internExpr(first, nenv)
        val rest_expr = internExprList(rest, nenv)
    in
        first_expr::rest_expr
    end

(*
 *  internType() takes a type and a name environment and returns the
 *   internalized version of the type.
 *)
and internType(UnevaluatedClassType{class_name=clName, is_owner=owner}, nenv) =
(*
 *  When internalizing a class, we replace the name (string) with the actual
 *  representation of the class.
 *)
    let
        val clVal = (lookup_nbinding(nenv, clName))
            handle NotFound => raise InternError("Class " ^ clName ^
                " not bound.")

        fun getTypeFromValue(cl) =
            case cl of
                ntClass(clOp) => ClassInstanceType{class=clOp, is_owner=owner}
              | _ => raise InternError("Class " ^ clName ^
                " is not a valid class.")
    in
        getTypeFromValue (clVal)
    end

  | internType(TupleType(tTypeList), nenv) =
(*
 *  Internalize the list of types.
 *)
    TupleType(internTypes(tTypeList,nenv))
  | internType(FunctionType(fType1, fType2), nenv) =
        FunctionType(internType(fType1,nenv),internType(fType2,nenv))
  | internType(ListType(lType), nenv) = ListType(internType(lType,nenv))
  | internType(RecordType(recordList), nenv) =
(*
 *  As with expressions, we alphabetize the list of records as we internalize
 *  them.
 *)
    let
        fun alphabetizeList(nil, newList) = newList
          | alphabetizeList(first :: rest, newList) =
            let
                fun alphaInsert(lastElt as {tident, ttype}, beforeList, nil) =
                    beforeList @ lastElt :: nil
                  | alphaInsert(elt as {tident, ttype}, beforeList,
                        (afterHead as {tident=aId, ttype=aType}
                        :{tident:string, ttype:Type})::afterRest) =
                    if (tident < aId) then
                    let
                        val newAfterList = afterHead :: afterRest
                        val newBeforeList = beforeList @ [elt]
                    in
                        newBeforeList @ newAfterList
                    end

                    else alphaInsert(elt, beforeList@[afterHead], afterRest)

                val insertedList = alphaInsert(first, nil, newList)
            in
                alphabetizeList(rest, insertedList)
            end

        val alphaList = alphabetizeList(recordList, nil)

        val tempRecFun = map(fn({tident=tid, ttype=t}) =>
                ({tident=tid, ttype=internType(t, nenv)}))

        val newRecordList = tempRecFun(alphaList)
    in
        RecordType(newRecordList)
    end

(*
 *  Basic types (RealType, StringType, IntegerType...) require no changes.
 *)
  | internType(otherType, nenv) = otherType

and internTypes(nil, nenv) = nil
  | internTypes(head::tail, nenv) = 
        internType(head, nenv)::internTypes(tail, nenv)

(*
 *  Takes a type option, and internalizes the option if it is SOME.
 *)
and internTypeOption(SOME(t), nenv) = SOME(internType(t, nenv))
  | internTypeOption(NONE, nenv) = NONE

(*
 *  Internalizes a pattern.  Returns a the internalized pattern and any new name
 *  envirnoment bindings.
 *)
and internPat(VarPattern(var_string), nenv):(PatternIntern * NEnvironment) = 
(*
 *  For every new var pattern, we add a binding saying that it isn't a class
 *  binding.
 *)
    let
        val new_nenv = new_nbinding(var_string, ntNotAClass)
    in
        (VarPatternIntern(var_string), new_nenv)
    end

  | internPat(WildcardPattern, nenv) =
        (WildcardPatternIntern, empty_nenvironment)
  | internPat(IntPattern(x), nenv) = (IntPatternIntern(x), empty_nenvironment)
  | internPat(RealPattern(x), nenv) = (RealPatternIntern(x), empty_nenvironment)
  | internPat(StringPattern(x), nenv) =
        (StringPatternIntern(x), empty_nenvironment)
  | internPat(BoolPattern(x), nenv) = (BoolPatternIntern(x), empty_nenvironment)
  | internPat(NilPattern, nenv) = (NilPatternIntern, empty_nenvironment)
  | internPat(TuplePattern(patList), nenv) = 
    let
        val (newPatList, pat_nenv) = internPats(patList, nenv)
    in
        (TuplePatternIntern(newPatList), pat_nenv)
    end
  | internPat(RecordPattern(recordList), nenv) = 
(*
 *  We must alphabetize the records as we internalize them.
 *)
    let
        fun alphabetizeList(nil, newList) = newList
          | alphabetizeList(first :: rest, newList) =
            let

                fun alphaInsert(lastElt as {ident, pattern}, beforeList, nil) =
                    beforeList @ lastElt :: nil
                  | alphaInsert(elt as {ident, pattern}, beforeList,
                        (afterHead as {ident=aId, pattern=aPat}
                        :{ident:string, pattern:Pattern})::afterRest) =
                    if (ident < aId) then
                    let
                        val newAfterList = afterHead :: afterRest
                        val newBeforeList = beforeList @ [elt]
                    in
                        newBeforeList @ newAfterList
                    end

                    else alphaInsert(elt, beforeList@[afterHead], afterRest)

                val insertedList = alphaInsert(first, nil, newList)
            in
                alphabetizeList(rest, insertedList)
            end

        val alphaList = alphabetizeList(recordList, nil)

        fun tmp({ident=tid, pattern=p}) = 
            let
                val (new_pat, new_nenv) = internPat(p, nenv)
            in
                ({ident=tid, pattern=new_pat}, new_nenv)
            end

        val tempRecFun = map(tmp)

        val newRecordAndNenvList = tempRecFun(alphaList)

        fun splitList(nil, list1, list2) = (list1, list2)
          | splitList((record, temp_nenv)::rest, list1, list2) =
                splitList(rest, list1@[record], list2@[temp_nenv])

        val (newRecordList, nenv_list) = splitList(newRecordAndNenvList, nil, nil)

        fun combineNenvList(nil) = empty_nenvironment
          | combineNenvList(first::rest) = combine_nenvs(first, combineNenvList(rest))

        val new_nenv = combineNenvList(nenv_list)

    in
        (RecordPatternIntern(newRecordList), new_nenv)
    end
  
  | internPat(ConsPattern({head=h, tail=t}), nenv) = 
    let
        val (head_pat, head_nenv) = internPat(h, nenv)
        val (tail_pat, tail_nenv) = internPat(t, nenv)
        val new_nenv = combine_nenvs(head_nenv, tail_nenv)
    in
        (ConsPatternIntern({head=head_pat, tail=tail_pat}), new_nenv)
    end
 
  | internPat(ClassPattern({class=cl_name:string,rep=cl_rep:Pattern}), nenv) =
    let 
        val (internRep, rep_nenv) = internPat(cl_rep, nenv)

        val ntVal:NenvValue = (lookup_nbinding(nenv, cl_name) 
		 handle NotFound => raise InternError("Class name " ^ cl_name ^
                        " not found in name env."))
	fun getClassFromValue(value) =
	    case value of
		ntClass(cl) => cl
	      | _ => raise InternError("ID is not a Class")
        
        val clValue:Class = getClassFromValue(ntVal)
    in
	(ClassPatternIntern({class=clValue, rep=internRep}), rep_nenv)
    end

(*
 *  interPats() takes a list of patterns and internalizes each of them.
 *)
and internPats(pat::pats, nenv) = 
    let
        val (head_pat, head_nenv) = internPat(pat, nenv)
        val (tail_pat, tail_nenv) = internPats(pats, nenv)
    in
        (head_pat::tail_pat, combine_nenvs(head_nenv, tail_nenv))
    end

  | internPats(nil, nenv) = (nil, empty_nenvironment)
