(*
 *  This is the main file that defines and invokes the read-eval-print loop
 *  function.
 *)

(*
 *  A helper to print out a welcome message when starting up the interpreter.
 *)
fun printWelcome() =
    let
        val infoText = "\nWelcome to MiniEML!\n" ^
                       "Enter the input declaration to be evaluated.\n" ^
                       "Enter a blank line to exit.\n"
    in
        print(infoText)
    end

(*
 *  An exception to identify when the input is empty.
 *)
exception EmptyInput

(*
 *  The input that the parser will read from.  Defaults to the console.  Can be
 *  changed to file.
 *)
val initialEMLenv = {instream = TextIO.stdIn, fromFile = 0} :: nil 

(*
 *  A function to read in the top-level decl or expr to evaluate, possibly over
 *  multiple lines.  raises EmptyInput if the input was blank.
 *)
fun readInput({instream, fromFile}):string =
    let
        val initialPrompt  = "MiniEML - "
	val continuePrompt = "MiniEML = "

(*
 *      A helper to read a single line of input, possibly extending previous lines.
 *)
	fun readLine(prompt:string, previousInput:string):string =
            let    
(*
 *              Print this line's prompt (only if not reading from a file).
 *)
		val _ = if (fromFile = 0) then print(prompt)
                        else print("")

(*
 *              Read a line of input.
 *)
		val inputString:string = TextIO.inputLine(instream)

(*
 *              Concatenate with input from previous lines
 *)
		val fullInputString:string = previousInput ^ inputString

(*
 *              Helper to check whether the string ends with a ";".
 *)
		fun ends_with_semicolon(s:string):bool =
		    let fun semicolon_helper(s:string, pos:int):bool =
			    if pos < 0 then raise EmptyInput
			    else
				let val last = String.sub(s, pos) 
				in
				    if last = #";" then true
				    else if last = #" " orelse 
					    last = #"\t" orelse 
					    last = #"\n" then
					 (* skip past whitespace char
					    and retry *)
					 semicolon_helper(s, pos-1)
		                    else false
		                end
		    in semicolon_helper(s, size(s)-1) end
	    in
		if ends_with_semicolon(fullInputString) then
(*
 *                  We're done with input; return the full string we've built so far.
 *)
		    fullInputString
	        else
(*
 *                  Continue reading another line of input.
 *)
		    readLine(continuePrompt, fullInputString)
	    end
    in
        readLine(initialPrompt, "")
    end


(* The main read-eval-print loop.  This is the function to invoke to
start the interpreter. *)

fun repl() =
    let fun loop(nenv:NEnvironment, tenv:tEnvironment, itcenv:itcEnvironment,
                env:Environment, EMLenv) =
        let
(*
 *          Read the input.  
 *)
            val curEMLenv = hd EMLenv
            val restEMLenv = tl EMLenv

            val inputString:string = readInput(curEMLenv)

(*
 *          Parse the input into an abstract syntax tree.
 *)
	    val ast:TopLevel = parse(inputString)

(*
 *          If we're reading from a file, check to see if we will be able to
 *          a line next time.  If not, we set the stream back to standard in.
 *)
            fun checkInputStream() =
            if (#fromFile curEMLenv = 1) then
            let
                val canRead = TextIO.lookahead(#instream curEMLenv)
            in
                case canRead of
                SOME(c) => EMLenv
              | NONE => restEMLenv
            end
            else EMLenv

            val tempEMLenv = checkInputStream()

(*
 *          Evaluate the parsed input, producing a set of new bindings.
 *)
            val (nbindings, tbindings, new_itcenv, bindings, newEMLenv) =
                ice(ast, nenv, tenv, itcenv, env, tempEMLenv)

(*
 *          Print the result bindings.
 *)
            val _ = emlPrint("New name env bindings:\n" ^ printNEnv(nbindings))
            val _ = emlPrint("New type env bindings:\n" ^ printtEnv(tbindings))
            val _ = emlPrint("Current itcenv:\n" ^ printitcEnv(new_itcenv))
            val _ = emlPrint("New env bindings:\n")
            val _ = print(printEnv(bindings)) (* always printed *)

(*
 *          Add in the new bindings, shadowing the previous ones.
 *)
            val new_nenv:NEnvironment = combine_nenvs(nenv, nbindings)
            val new_tenv:tEnvironment = combine_tenvs(tenv, tbindings)
            val new_env:Environment = combine_envs(env, bindings)

        in
       	    loop(new_nenv, new_tenv, new_itcenv, new_env, newEMLenv)
        end

	handle	       
	    (* handle any errors that might have happened in the any of
	       the steps above *)

	    ParseError(msg:string) =>
	        (* there was a syntax error.  print the error message,
		   then continue the repl loop *)
		(print("syntax error: " ^ msg ^ "\n");
		 loop(nenv, tenv, itcenv, env, EMLenv))
          
          | InternError(msg:string) =>
                (* there was an error during internalization. *)
                (print("Internalization Error: " ^ msg ^ "\n");
                 loop(nenv, tenv, itcenv, env, EMLenv))

          | TypeError(msg:string) =>
            (*
             *  There was an error during type checking.
             *)
            (print("Type checking error: " ^ msg ^ "\n");
                loop(nenv, tenv, itcenv, env, EMLenv))
	  | EvalError(msg:string) =>
	        (* there was an error during evaluation.  print the
		   error message, then continue the repl loop *)
		(print("Evaluation Error: " ^ msg ^ "\n");
		 loop(nenv, tenv, itcenv, env, EMLenv))
(*          | EmptyInput => loop(nenv, tenv, itcenv, env, EMLenv)*)
	  
    in
       (printWelcome();
        (* start off the repl loop with empty environment *)
        loop(empty_nenvironment, empty_tenvironment, empty_itcenvironment,
                empty_environment, initialEMLenv))
    end

    handle EmptyInput =>
        (* we're done! *)
         "Quiting MiniEML!"
