(*
 *  emlstructs.sml: This file contains the types and datatypes for EML.
 *  
 *  A commented version of the AST is available at the bottom of this file.
 *)


(* 
 *  Used for comparing patterns.
 *)
datatype Specificity =
          SameSpecificity
        | Incomperable 
        | FirstIsMoreSpecific
        | SecondIsMoreSpecific

datatype TypeComparison =
          NotSameType
        | SameType

(*
 *  ClassOwnerType is used in types to establish whether the class type has
 *  been marked as an owner or not.
 *)
datatype ClassOwnerType =
	  OwnerClass
	| NonOwnerClass

(*
 *  ClassMode represents the diffent kinds of classes that are recognized. 
 *)
datatype ClassMode =
          Interface
        | Abstract
        | Concrete

(*
 *  Class is the internal representation of a class.
 *)
datatype Class =
          Class of {name: string, id: int, mode: ClassMode,
                    superclass: Class option, interfaces: Class list,
                    dType: Type option}

(*
 *  NenvValue is used in the name environment (in the internalization phase).
 *  It simply tells if a string is a class or not.  If it is a class, the
 *  class is also present, so a substitution from the string to the class con
 *  be done.
 *)
and NenvValue =
          ntNotAClass
        | ntClass of Class

(*
 *  Value is used in the regular environment (in the evaluation phase).
 *  It contains links between an identifier (string) and the value it 
 *  represents.
 *)
and Value =
          IntValue of int
        | StringValue of string
        | RealValue of real
        | BoolValue of bool
	| TupleValue of Value list
        | ListValue of Value list
        | RecordValue of {ident:string, value:Value} list
        | FunValue of {name: string, cases: {formal:PatternIntern, 
                       body:ExprIntern} list, env: Environment, dType:Type}
        | ClassObject of {class:Class, rep:Value}
        | GenericFunValue of {mdList:Method list ref, dType:Type}
        | ClassValue of Class

(* 
 *  The controlling level datatype.  TopLevelValDecl and TopLevelOODecl are
 *  passed to the internalization phase.  TopLevelValDeclIntern and
 *  TopLevelOODeclExpr are passed to the type check and evaluation phase.
 *  TopLevelSysCall is caught before the ICE process, and handled without
 *  touching the ICE process.
 *)
and TopLevel =
          TopLevelValDecl of vDecl
        | TopLevelOODecl of ooDecl list
        | TopLevelValDeclIntern of vDeclIntern
        | TopLevelOODeclIntern of ooDeclIntern list
        | TopLevelSysCall of SysCallID

(*
 *  TopLevelValDecl are only val declarations.
 *
 *  Examples:
 *  val (x, y) = (3, "me")
 *  val clInst:Colin = Colin(3);
 *)
and vDecl =
          ValDecl of {lhs:Pattern, lhs_type:Type, rhs:Expr}
        | UnTypedValDecl of {lhs:Pattern, rhs:Expr}

(*
 *  TopLevelOODecl (Object Oriented) are all declarations involving object
 *  oriented stuff.
 *
 *  Examples:
 *  FunDecl => fun funName: int -> string (3) = "three"
 *                                      | (x) = "not three!";
 *  ClassDecl => class SubCl extends SupCl implements Int1, Int2 of (int * int)
 *  InterfaceDecl => interface Int2 extends Int1;
 *  MessageDecl => message gfName : int * int -> int;
 *  MethodDecl => method gfName(3, 5) = 8;
 *)
and ooDecl =
          FunDecl of {name:string, decType:Type,
                      fcases:{formal:Pattern, body:Expr} list}
	| ClassDecl of {name:string, mode:ClassMode, superclass:string option,
		 interfaces:string list, dType:Type option}
	| InterfaceDecl of {name:string, interfaces:string list}
	| MessageDecl of {name:string, dType:Type}
	| MethodDecl of {name:string, formal:Pattern, body:Expr}

(*
 *  Commands for EML to handle.  
 *
 *  SysCallUse tells EML to run a file.
 *  SysCallVerbose tells EML to turn on or off run verbose.
 *)
and SysCallID =
          SysCallUse of string
        | SysCallVerbose of int
    
(* 
 *  Expr is passed in from the parser to the top level control.  It is then
 *  converted to an internal format (represented by ExprIntern).
 *)
and Expr =
          VarExpr of string
        | IntExpr of int
        | RealExpr of real
        | StringExpr of string
        | BoolExpr of bool
        | NilExpr of Type
	| TupleExpr of Expr list
        | RecordExpr of {eident:string, evalue:Expr} list
	| UnopExpr of {operator:string, arg:Expr}
	| BinopExpr of {arg1:Expr, operator:string, arg2:Expr}
	| CallExpr of {function:Expr, arg:Expr}
	| IfExpr of {testExpr:Expr, thenExpr:Expr, elseExpr:Expr}
	| LetExpr of {decls:vDecl list, body:Expr}

(* 
 *  Pattern is passed in from the parser to the top level control.  It is then
 *  converted to an internal format (represented by PatternIntern).
 *)
and Pattern =
          VarPattern of string
	| WildcardPattern
        | IntPattern of int
        | RealPattern of real
        | StringPattern of string
        | BoolPattern of bool
        | NilPattern
	| TuplePattern of Pattern list
        | RecordPattern of {ident:string, pattern:Pattern} list
	| ConsPattern of {head:Pattern, tail:Pattern}
	| ClassPattern of {class:string, rep:Pattern}

(* 
 *  TopLevelValDeclIntern is returned from the internalization phase (for a
 *  TopLevelValDeclIntern).  This is then passed to the type check and
 *  evaluation phase.
 *)
and vDeclIntern = 
          ValDeclIntern of {lhs:PatternIntern, lhs_type:Type, rhs:ExprIntern}
        | UnTypedValDeclIntern of {lhs:PatternIntern, rhs:ExprIntern}

(* 
 *  TopLevelOODeclIntern is returned from the internalization phase (for a
 *  TopLevelOODeclIntern).  This is then passed to the type check and
 *  evaluation phase.
 *)
and ooDeclIntern =
          FunDeclIntern of {name:string, iType:Type, 
                            fcases:{formal:PatternIntern, body:ExprIntern} list}
	| ClassDeclIntern of Class
	| InterfaceDeclIntern of Class
	| MessageDeclIntern of {name:string, iType:Type}
	| MethodDeclIntern of {name:string, formal:PatternIntern,
                                body:ExprIntern}

(* 
 *  ExprIntern is returned from the internalization phase.  This is then passed
 *  to the evaluation phase.
 *)
and ExprIntern =
          VarExprIntern of string
        | IntExprIntern of int
        | RealExprIntern of real
        | StringExprIntern of string
        | BoolExprIntern of bool
        | NilExprIntern of Type
	| TupleExprIntern of ExprIntern list
        | RecordExprIntern of {eident:string, evalue:ExprIntern} list
	| UnopExprIntern of {operator:string, arg:ExprIntern}
	| BinopExprIntern of {arg1:ExprIntern, operator:string, arg2:ExprIntern}
	| CallExprIntern of {function:ExprIntern, arg:ExprIntern}
	| IfExprIntern of {testExpr:ExprIntern, thenExpr:ExprIntern,
                           elseExpr:ExprIntern}
	| LetExprIntern of {decls:vDeclIntern list, body:ExprIntern}
        | ClassExprIntern of Class

(* 
 *  PatternIntern is returned from the internalization phase.  This is then
 *  passed to the evaluation phase.
 *)
and PatternIntern =
          VarPatternIntern of string
	| WildcardPatternIntern
        | IntPatternIntern of int
        | RealPatternIntern of real
        | StringPatternIntern of string
        | BoolPatternIntern of bool
        | NilPatternIntern
	| TuplePatternIntern of PatternIntern list
        | RecordPatternIntern of {ident:string, pattern:PatternIntern} list
	| ConsPatternIntern of {head:PatternIntern, tail:PatternIntern}
	| ClassPatternIntern of {class:Class, rep:PatternIntern}

(*
 * Type represents the types recognized.
 *)
and Type =
           IntegerType 
         | RealType
         | BooleanType
         | StringType
         | TupleType of Type list
         | ListType of Type
         | UnevaluatedClassType of {class_name:string, is_owner:ClassOwnerType}
         | ClassInstanceType of {class:Class, is_owner:ClassOwnerType}
         | RecordType of {tident:string, ttype:Type} list
         | FunctionType of (Type * Type)

(* 
 *  Environment is used to store links between identifiers (strings) and the
 *  values they represent.
 *)
withtype Environment = (string, Value) Assoc_List.assoc_list

(*
 *  NEnvironment is used to stron links between identifiers (strings) and
 *  whether or not that identifier is a class or not.  If it is a class,
 *  the class representation is stored as well.
 *)
and NEnvironment = (string, NenvValue) Assoc_List.assoc_list

(*
 *  Method is what makes up a generic function.
 *)
and Method = {formal:PatternIntern, body:ExprIntern, env:Environment}


(*
----------------------------------------------------------------
*This is currently out of date.  It will be updated eventually.*
----------------------------------------------------------------

The file defines the AST datatype that represents programs that can be
read in by the MiniML interpreter.  The syntax that is supported is as follows:

TopLevel ::= Decl | Expr

Decl ::= "val" Pattern "=" Expr
       | "fun" {Id Pattern "=" Expr}/"|"
       | <"abstract"> "class" Id <"extends" Id> <"implements" {Id}/" ">
       | "interface" Id <extends {Id}/" ">
       | "message" Id
       | "method" Id Pattern "=" Expr

Expr ::= Id				(* a variable reference *)
       | Int | String | "true" | "false" | "nil"	(* literal constants *)
       | "(" ")"			(* the unit value *)
       | "(" Expr "," Expr {"," Expr} ")"     (* a tuple of 2 or more exprs *)
       | Unop Expr | Expr Binop Expr	(* unary and binary expressions *)
       | Id Expr			(* a function call *)
       | "if" Expr "then" Expr "else" Expr	(* an if expression *)
       | "let" {Decl} "in" Expr "end"	(* a let expression *)
       | "(" Expr ")"			(* a simple parenthesized expression *)

Unop ::= "~" | "not"

Binop ::= "+" | "-" | "*" | "div" | "mod" 
        | "=" | "<>" | "<" | "<=" | ">=" | ">"
        | "^" | "::"

Pattern ::= Id | "_"			(* variable and wildcard patterns *)
          | Int | String | "true" | "false" | "nil"	(* literal patterns *)
          | "(" ")"			(* the unit pattern *)
          | "(" Pattern "," Pattern {"," Pattern} ")"	(* a tuple pattern *)
	  | Pattern "::" Pattern	(* a list pattern *)
          | "(" Pattern ")"		(* a simple parenthesized pattern *)
          | Id Pattern                  (* a class pattern *)
*)
