(*
 *  emlstructs_print.sml: A pretty print function for the types and datatypes
 *  defined in emlstructs.sml.
 *)

fun printTopLevel(TopLevelValDecl(vdecl)) = printVDecl(vdecl)
  | printTopLevel(TopLevelOODecl(oodecl)) = printOODeclList(oodecl)
  | printTopLevel(TopLevelValDeclIntern(vdecl)) = printVDeclIntern(vdecl)
  | printTopLevel(TopLevelOODeclIntern(oodecl)) = printOODeclInternList(oodecl)
  | printTopLevel(TopLevelSysCall(sysCallID)) = "Top Level System Call"

and printVDecl(ValDecl{lhs, lhs_type, rhs}) = 
        "val " ^ printPattern(lhs) ^ " : " ^ printType(lhs_type) ^ " = " ^
         printExpr(rhs)
  | printVDecl(UnTypedValDecl{lhs, rhs}) = 
        "val " ^ printPattern(lhs) ^ " = " ^ printExpr(rhs)

and printOODecl(FunDecl({name, decType, fcases})) = 
        "fun " ^ name ^ " : " ^ printType(decType) ^ printCaseList(fcases)
  | printOODecl(ClassDecl({name, mode, superclass, interfaces, dType})) =
        printClassMode(mode) ^ " " ^ name ^ " : " ^ printTypeOpt(dType) ^
        printSClassOption(superclass) ^ printIClassList(interfaces)
  | printOODecl(InterfaceDecl({name, interfaces})) =
        "interface " ^ name ^ printIClassList(interfaces)
  | printOODecl(MessageDecl({name,dType})) = 
        "message " ^ name ^ " : " ^ printType(dType)
  | printOODecl(MethodDecl({name, formal, body})) = 
        "method " ^ name ^ printPattern(formal) ^ " = " ^ printExpr(body)

and printExpr(VarExpr(name)) = name
  | printExpr(IntExpr(i)) = Int.toString(i)
  | printExpr(RealExpr(r)) = Real.toString(r)
  | printExpr(StringExpr(s)) = "\"" ^ s ^ "\""
  | printExpr(BoolExpr(true)) = "true"
  | printExpr(BoolExpr(false)) = "false"
  | printExpr(NilExpr(decType)) = "nil : " ^ printType(decType)
  | printExpr(TupleExpr(exprs)) = "(" ^ printExprList(exprs) ^ ")"
  | printExpr(RecordExpr(recs)) = "{" ^ printRecExprList(recs) ^ "}"
  | printExpr(UnopExpr{operator,arg}) = 
        "(" ^ operator ^ " " ^ printExpr(arg) ^ ")"
  | printExpr(BinopExpr{arg1,operator,arg2}) = 
        "(" ^ printExpr(arg1) ^ " " ^ operator ^ " " ^ printExpr(arg2) ^ ")"
  | printExpr(CallExpr{function,arg}) = 
        printExpr(function) ^ " (" ^ printExpr(arg) ^ ")"
  | printExpr(IfExpr{testExpr,thenExpr,elseExpr}) = 
        "if " ^ printExpr(testExpr) ^ " then " ^ printExpr(thenExpr) ^ " else "
        ^ printExpr(elseExpr)
  | printExpr(LetExpr{decls,body}) = 
        "let " ^ printVDeclList(decls) ^ " in " ^ printExpr(body) ^ " end"

and printRecExprList(nil) = ""
  | printRecExprList([{eident, evalue}]) = eident ^ " = " ^ printExpr(evalue)
  | printRecExprList({eident, evalue}::rest) = eident ^ " = " ^
        printExpr(evalue) ^ ", " ^ printRecExprList(rest)

and printPattern(VarPattern(name)) = name
  | printPattern(WildcardPattern) = "_"
  | printPattern(IntPattern(i)) = Int.toString(i)
  | printPattern(RealPattern(r)) = Real.toString(r)
  | printPattern(StringPattern(s)) = "\"" ^ s ^ "\""
  | printPattern(BoolPattern(true)) = "true"
  | printPattern(BoolPattern(false)) = "false"
  | printPattern(NilPattern) = "nil"
  | printPattern(TuplePattern(pats)) = "(" ^ printPatternList(pats) ^ ")"
  | printPattern(RecordPattern(recordList)) = 
        "{" ^ printRecordList(recordList) ^ "}"
  | printPattern(ConsPattern{head, tail}) = 
        "(" ^ printPattern(head) ^ " :: " ^ printPattern(tail) ^ ")"
  | printPattern(ClassPattern({class, rep})) = 
        "class " ^ class ^ " : " ^ printPattern(rep)

and printVDeclList([]) = ""
  | printVDeclList([decl]) = printVDecl(decl)
  | printVDeclList(decl::decl2::decls) = 
        printVDecl(decl) ^ "\n\t" ^ printVDeclList(decl2::decls)

and printOODeclList([]) = ""
  | printOODeclList([decl]) = printOODecl(decl)
  | printOODeclList(decl::decl2::decls) = 
        printOODecl(decl) ^ "\n\t" ^ printOODeclList(decl2::decls)

and printExprList([]) = ""
  | printExprList([expr]) = printExpr(expr)
  | printExprList(expr::expr2::exprs) = 
        printExpr(expr) ^ ", " ^ printExprList(expr2::exprs)

and printPatternList([]) = ""
  | printPatternList([pattern]) = printPattern(pattern)
  | printPatternList(pattern::pattern2::patterns) =
	printPattern(pattern) ^ ", " ^ printPatternList(pattern2::patterns)

and printRecordList(nil) = ""
  | printRecordList([{ident, pattern}]) = ident ^ " = " ^ printPattern(pattern)
  | printRecordList({ident, pattern} :: restRecords) = ident ^ " = " ^
        printPattern(pattern) ^ ", " ^ printRecordList(restRecords)

(* 
 *  Internal printing.
 *)
and printVDeclIntern(ValDeclIntern{lhs, lhs_type, rhs}) = 
        "val " ^ printPatternIntern(lhs) ^ " : " ^ printType(lhs_type) ^
         " = " ^ printExprIntern(rhs)
  | printVDeclIntern(UnTypedValDeclIntern{lhs, rhs}) =
        "val " ^ printPatternIntern(lhs) ^ " = " ^ printExprIntern(rhs)

and printOODeclIntern(FunDeclIntern({name,iType,fcases})) = 
        "fun " ^ name ^ " : " ^ printType(iType) ^  printCaseInternList(fcases)
  | printOODeclIntern(ClassDeclIntern(Class{name, id, mode, superclass,
        interfaces, dType})) =
        printClassMode(mode) ^ " " ^ name ^ " [id " ^ Int.toString(id) ^ "] : " ^ 
        printTypeOpt(dType) ^ printSClassName(superclass) ^ 
        printIClassNames(interfaces)
  | printOODeclIntern(InterfaceDeclIntern(Class{name, id, mode, superclass,
        interfaces, dType})) =
        "interface " ^ name ^ " [id " ^ Int.toString(id) ^ "]" ^ 
        printIClassNames(interfaces)
  | printOODeclIntern(MessageDeclIntern({name,iType})) = 
        "message " ^ name ^ " : " ^ printType(iType)
  | printOODeclIntern(MethodDeclIntern({name, formal, body})) = 
        "method " ^ name ^ " " ^ printPatternIntern(formal) ^ 
        printExprIntern(body)

and printExprIntern(VarExprIntern(name)) = name
  | printExprIntern(IntExprIntern(i)) = Int.toString(i)
  | printExprIntern(RealExprIntern(r)) = Real.toString(r)
  | printExprIntern(StringExprIntern(s)) = "\"" ^ s ^ "\""
  | printExprIntern(BoolExprIntern(true)) = "true"
  | printExprIntern(BoolExprIntern(false)) = "false"
  | printExprIntern(NilExprIntern(decType)) = "nil : " ^ printType(decType)
  | printExprIntern(TupleExprIntern(exprs)) = 
        "(" ^ printExprInternList(exprs) ^ ")"
  | printExprIntern(RecordExprIntern(recs)) =
        "{" ^ printRecExprInternList(recs) ^ "}"
  | printExprIntern(UnopExprIntern{operator,arg}) = 
        "(" ^ operator ^ " " ^ printExprIntern(arg) ^ ")"
  | printExprIntern(BinopExprIntern{arg1,operator,arg2}) = 
        "(" ^ printExprIntern(arg1) ^ " " ^ operator ^ " " ^ 
        printExprIntern(arg2) ^ ")"
  | printExprIntern(CallExprIntern{function,arg}) = 
        printExprIntern(function) ^ " " ^ printExprIntern(arg)
  | printExprIntern(IfExprIntern{testExpr,thenExpr,elseExpr}) = 
        "if " ^ printExprIntern(testExpr) ^ " then " ^ printExprIntern(thenExpr)
        ^ " else " ^ printExprIntern(elseExpr)
  | printExprIntern(LetExprIntern{decls,body}) = 
        "let " ^ printVDeclInternList(decls) ^ " in " ^  printExprIntern(body) ^
        " end"
  | printExprIntern(ClassExprIntern(Class{name, id, mode, superclass, 
        interfaces, dType})) = 
        printClassMode(mode) ^ " " ^ name ^ " [id " ^ Int.toString(id) ^ "] : " ^
        printTypeOpt(dType) ^ printSClassName(superclass) ^ 
        printIClassNames(interfaces)

and printRecExprInternList(nil) = ""
  | printRecExprInternList([{eident, evalue}]) = eident ^ " = " ^ printExprIntern(evalue)
  | printRecExprInternList({eident, evalue}::rest) = eident ^ " = " ^
        printExprIntern(evalue) ^ ", " ^ printRecExprInternList(rest)

and printPatternIntern(VarPatternIntern(name)) = name
  | printPatternIntern(WildcardPatternIntern) = "_"
  | printPatternIntern(IntPatternIntern(i)) = Int.toString(i)
  | printPatternIntern(RealPatternIntern(r)) = Real.toString(r)
  | printPatternIntern(StringPatternIntern(s)) = "\"" ^ s ^ "\""
  | printPatternIntern(BoolPatternIntern(true)) = "true"
  | printPatternIntern(BoolPatternIntern(false)) = "false"
  | printPatternIntern(NilPatternIntern) = "nil"
  | printPatternIntern(TuplePatternIntern(patterns)) = 
        "(" ^ printPatternInternList(patterns) ^ ")"
  | printPatternIntern(RecordPatternIntern(records)) =
        "{" ^ printRecordInternList(records) ^ "}"
  | printPatternIntern(ConsPatternIntern{head, tail}) = 
        "(" ^ printPatternIntern(head) ^ " :: " ^ printPatternIntern(tail) ^
        ")"
  | printPatternIntern(ClassPatternIntern({class=Class{name, id, mode, 
        superclass, interfaces, dType}, rep})) =
        printClassMode(mode) ^ " " ^ name ^ " [id " ^ Int.toString(id) ^ "] : " ^
        printTypeOpt(dType) ^ printSClassName(superclass) ^ 
        printIClassNames(interfaces) ^ " as " ^ printPatternIntern(rep)

and printVDeclInternList([]) = ""
  | printVDeclInternList([decl]) = printVDeclIntern(decl)
  | printVDeclInternList(decl::decl2::decls) = 
        printVDeclIntern(decl) ^ "\n\t" ^ printVDeclInternList(decl2::decls)

and printOODeclInternList([]) = ""
  | printOODeclInternList([decl]) = printOODeclIntern(decl)
  | printOODeclInternList(decl::decl2::decls) = 
        printOODeclIntern(decl) ^ "\n\t" ^ printOODeclInternList(decl2::decls)

and printExprInternList([]) = ""
  | printExprInternList([expr]) = printExprIntern(expr)
  | printExprInternList(expr::expr2::exprs) = 
        printExprIntern(expr) ^ ", " ^ printExprInternList(expr2::exprs)

and printPatternInternList([]) = ""
  | printPatternInternList([pattern]) = printPatternIntern(pattern)
  | printPatternInternList(pattern::pattern2::patterns) =
        printPatternIntern(pattern) ^ ", " ^ 
        printPatternInternList(pattern2::patterns)

and printRecordInternList(nil) = ""
  | printRecordInternList([{ident, pattern}]) = 
        ident ^ " = " ^ printPatternIntern(pattern)
  | printRecordInternList({ident, pattern} :: restRecords) = ident ^ " = " ^
        printPatternIntern(pattern) ^ ", " ^ printRecordInternList(restRecords)

(*
 *  Printing for types.
 *)

and printType(IntegerType) = "int"
  | printType(RealType) = "real"
  | printType(BooleanType) = "bool"
  | printType(StringType) = "string"
  | printType(TupleType(types)) = "(" ^ printTypeList(types) ^ ")"
  | printType(ListType(l_type)) = printType(l_type) ^ " list"
  | printType(UnevaluatedClassType{class_name=name, is_owner=NonOwnerClass}) =
        "unevaluated class " ^ name
  | printType(UnevaluatedClassType{class_name=name, is_owner=OwnerClass}) =
        "unevaluated class " ^ "# " ^ name
  | printType(ClassInstanceType{class=cl, is_owner=NonOwnerClass}) =
        printClassName(cl) ^ " instance"
  | printType(ClassInstanceType{class=cl, is_owner=OwnerClass}) =
        printClassName(cl) ^ " instance (marked as owner)"
  | printType(RecordType(recordTypeList)) = 
        "{" ^ printRecordTypeList(recordTypeList) ^ "}"
  | printType(FunctionType(arg_type, ret_type)) = printType(arg_type) ^ " -> " ^
                                                 printType(ret_type)

and printTypeList(nil) = "()"
  | printTypeList([head_type]) = printType(head_type)
  | printTypeList(head_type::next_type::rest_types) = printType(head_type) ^ 
        " * " ^ printTypeList(next_type::rest_types)

and printTypeOpt(SOME(t)) = printType(t)
  | printTypeOpt(NONE) = ""

and printRecordTypeList(nil) = ""
  | printRecordTypeList([{tident, ttype}]) = tident ^ ":" ^ printType(ttype)
  | printRecordTypeList({tident, ttype} :: restRecords) = tident ^ ":" ^
        printType(ttype) ^ ", " ^ printRecordTypeList(restRecords)

(* 
 *  Printing for Values and NenvValues
 *)
and printValue(IntValue(i)) = Int.toString(i)
  | printValue(RealValue(r)) = Real.toString(r)
  | printValue(StringValue(s)) = "\"" ^ s ^ "\""
  | printValue(BoolValue(true)) = "true"
  | printValue(BoolValue(false)) = "false"
  | printValue(TupleValue(values)) = 
        "(" ^ printValueList(values) ^")"
  | printValue(ListValue(values)) = "[" ^ printValueList(values) ^ "]"
  | printValue(RecordValue(recList)) = "{" ^ printRecordValueList(recList) ^ "}"
  | printValue(FunValue{name,cases,env,dType}) = 
        "fn " ^ name ^ " : " ^ printType(dType) ^ printValueCaseList(cases)
  | printValue(ClassObject{class=Class{name,id,mode,superclass,interfaces,
        dType},rep}) = 
        "class " ^ name ^ " as " ^ printValue(rep)
  | printValue(ClassValue(Class{name, id, mode, superclass, interfaces, 
        dType})) =
        "class " ^ name
  | printValue(GenericFunValue({mdList=msgList, dType=dt})) =
        case !msgList of
	    nil => "Empty Generic Function"
	  | _ => "Populated Generic Function"

and printValueCaseList([]) = ""
  | printValueCaseList([fcase]) = printValueCase(fcase)
  | printValueCaseList(fcase::fcase2::fcases) =
        printValueCase(fcase) ^ "\n | " ^ printValueCaseList(fcase2::fcases)

and printValueCase({formal,body}) = 
        printPatternIntern(formal) ^ " => " ^ printExprIntern(body)

and printValueList([]) = ""
  | printValueList([value]) = printValue(value)
  | printValueList(value::value2::values) = 
        printValue(value) ^ ", " ^ printValueList(value2::values)

and printRecordValueList([]) = ""
  | printRecordValueList([{ident, value}]) = ident ^ " = " ^ printValue(value)
  | printRecordValueList({ident, value} :: restRecs) =
        ident ^ " = " ^ printValue(value) ^ ", " ^ printRecordValueList(restRecs)

and printNenvValue(ntNotAClass) = "not a class"
  | printNenvValue(ntClass(_)) = "a class"

(*
 *  Helper functions for the above printing code.
 *)
and printCaseList([]) = ""
  | printCaseList([fcase]) = printCase(fcase)
  | printCaseList(fcase::fcase2::fcases) = 
        printCase(fcase) ^ "\n  | " ^ printCaseList(fcase2::fcases)

and printCase({formal,body}) = 
        " " ^ printPattern(formal) ^ " = " ^ printExpr(body)

and printCaseInternList([]) = ""
  | printCaseInternList([fcase]) = printCaseIntern(fcase)
  | printCaseInternList(fcase::fcase2::fcases) = 
        printCaseIntern(fcase) ^ "\n  | " ^ printCaseInternList(fcase2::fcases)

and printCaseIntern({formal,body}) = 
        " " ^ printPatternIntern(formal) ^ " = " ^ printExprIntern(body)

(* 
 *  Takes a string option and prints out the string if it exists, 
 *  otherwise it prints out nothing.
 *)
and printSClassName(NONE) = ""
  | printSClassName(SOME (Class({name,id,mode,superclass,interfaces,dType}))) = 
        " superclass " ^ name

and printSClassOption(NONE) = ""
  | printSClassOption(SOME cl_name) = " superclass " ^ cl_name

(* 
 *  Takes a list of strings and prints them out, separated by " ".
 *)
and printIClassNames(nil) = ""
  | printIClassNames(xs) = " implements " ^ printInterfaceNames(xs)

and printInterfaceNames(nil) = ""
  | printInterfaceNames(Class({name,id,mode,superclass,interfaces,
        dType})::iClasses) =
        name ^ " " ^ printInterfaceNames(iClasses)

and printIClassList(nil) = ""
  | printIClassList(xs) = 
    let
        fun printHelper(nil) = ""
          | printHelper(y::ys) = y ^ " " ^ printHelper(ys)
    in
        " implements " ^ printHelper(xs)
    end

(* 
 *  Prints out the type of class.
 *)
and printClassMode(Abstract) = "abstract class"
  | printClassMode(Concrete) = "class"
  | printClassMode(Interface) = "interface"

and printClassName(Class{name, id, mode, superclass, interfaces, dType}) = name

and printitcEnv(nil) = ""
  | printitcEnv({name=n, gblType=t, pats=ps}::rest) =
        n ^ "(" ^ printType(t) ^ "): " ^ printPatternInternList(ps) ^ "\n\n" ^
        printitcEnv(rest)
