(*
 *  emlTypeCheck.sml: This file is the second step in the ICE process.  It 
 *  takes an internalized top level and produces a new type environment.
 *  In the process, it also checks to see if the internalized top level is
 *  correctly typed.
 *)


(* 
 *  typeCheck() takes the top-level decl or expr, and the top-level type
 *  environment (i.e., the bindings of global variables to their types), and
 *  then type checks the top-level ast and produces a set of new type bindings
 *  which is returned to the caller.
 *)
fun typeCheck(ast:TopLevel, tenv:tEnvironment, itcenv, name_env) =
    let
        val infoText = "[Type checking " ^ printTopLevel(ast) ^ "]\n"
        val _ = emlPrint(infoText)
    in
        typeTopLevel(ast, tenv, itcenv, name_env)
    end

(* 
 *  typeTopLevel does the work of typeCheck().
 *)
and typeTopLevel(TopLevelValDeclIntern(decl), tenv, itcenv, name_env) = 
    let
        val new_tenv =  typeVDecl(decl, tenv)
    in
        (new_tenv, itcenv)
    end
  | typeTopLevel(TopLevelOODeclIntern(decl), tenv, itcenv, name_env) =
    let
        val (new_tenv, new_itcenv, menv) = typeOODeclList(decl, tenv, itcenv)

        (*
         *  Make sure that the new classes obey the ambiguity rules we're
         *  forcing.
         *)
        val _ = findOwnerAmbiguities(tenv, new_tenv, menv)
(*
 *      If itc() returns then the implementation side type checking has
 *      passed.  If itc() fails then it will raise an exception.
 *)
(*        val _ = itc(combine_tenvs(tenv, new_tenv), new_itcenv)*)
        val _ = emlPrint("Beginning ITC code.\n")
        val _ = itc(name_env, new_itcenv)
    in
        (new_tenv, new_itcenv)
    end
  | typeTopLevel(_) = raise TypeError("Type checking phase erronously recieved" ^
                                       " non-internalized top level.")

(*
 *  To type check a val decl, we type check the pattern and the expr, and 
 *  then check the equality of the types.
 *)
and typeVDecl(ValDeclIntern{lhs:PatternIntern, lhs_type, rhs:ExprIntern}, tenv) =
    let 
        val lhs_tenv = typePattern(lhs, lhs_type)
(*
 *  <colin> 11/4/2001 -- Should we need to extend the environment here?
 *)
        val extended_tenv = combine_tenvs(tenv, lhs_tenv)
        val rhs_type = typeExpr(rhs, extended_tenv)
        val firstAsSpecific  = typeFirstAsSpecific(rhs_type, lhs_type)
    in
        if (firstAsSpecific) then lhs_tenv
        else raise TypeError("RHS type " ^ printType(rhs_type) ^ " doesn't " ^
            "match LHS type " ^ printType(lhs_type) ^ ".")
    end

  | typeVDecl(UnTypedValDeclIntern{lhs:PatternIntern, rhs:ExprIntern}, tenv) =
    let
        val rhs_type = typeExpr(rhs, tenv)
        val lhs_tenv = typePattern(lhs, rhs_type)
    in
        lhs_tenv
    end

(*
 *  typeOODecl(): type checks an object-oriented declaration.  An object
 *      oriented declaration is not a val declaration.
 *
 *  Parameters: An internalized declaration to be type checked, a type
 *      environment, and a itc environment.
 *
 *  Return Values: Returns a new type environment containing the type bindings
 *      created in the declaration, the full new itc environment, and a method
 *      envrironment.  The method environment is either a (name, formal, body)
 *      option; SOME for method declarations and NONE for non-method 
 *      declarations.
 *
 *  Comments: The method environment is needed because we need to know which
 *      methods were created locally during the module type checking phase.
 *      Since new methds aren't included in the type environment (only
 *      messages are), we need to keep track of this information separately.
 *)
and typeOODecl(FunDeclIntern({name=internalName,iType=evaledType,fcases}),
        tenv, itcenv) =
    (*
     *  To type check a function, we first type check all the cases.  We then 
     *  take the type of the cases and compare it to the declared type.
     *)
    let
        fun getAndCheckFunParts(check_type) =
            if (isValidFunType(check_type)) then typeGetFunParts(check_type)
            else raise TypeError("Function " ^ internalName ^ " has type: " ^
                     printType(check_type) ^ " which is not a valid function "
                     ^ "type.")
                
        val (pat_type, act_body_type) = getAndCheckFunParts(evaledType)
        val _ = print("pat type is: " ^ printType(pat_type) ^ "\n")
        val _ = print("act body type is: " ^ printType(act_body_type) ^ "\n")
        (*
         *  Mismatching cases are caught in the typePattern() call, as they
         *  must match the declared type.
         *)
        fun typeCases([]) = pat_type
          | typeCases({formal, body} :: rest) =
            let
                val _ = print("checking formal: " ^ printPatternIntern(formal) ^
                        " and body: " ^ printExprIntern(body) ^ "\n")
                val fml_tenv = typePattern(formal, pat_type)
                val extended_tenv = combine_tenvs(tenv, fml_tenv)
                val bd_type = typeExpr(body, extended_tenv)
                val firstAsSpecific = typeFirstAsSpecific(bd_type, act_body_type)
            in
                if (firstAsSpecific) then typeCases(rest)
                else raise TypeError("Type of body: " ^ printType(bd_type) ^ 
                         " does not match declared return type: " ^ 
                         printType(act_body_type) ^ ".")
            end

        val cases_type = typeCases(fcases)
        val new_tenv = new_tbinding(internalName, evaledType)

    in
        (new_tenv, itcenv, NONE)
    end
  
  | typeOODecl(ClassDeclIntern(cl as Class{name, id, mode, superclass, interfaces,
        dType}), tenv, itcenv) =
    (*  
     *  To type classes, we can only need to create a new binding from the
     *  class name to the type.  We also check to make sure that the type is a
     *  valid class type.
     *)
    if (isValidClassTypeOpt(dType)) then 
        (*
         * COLIN: Is this what we want?  Do we just blindly set is owner to no?
         *)
        (new_tbinding(name, ClassInstanceType{class=cl, is_owner=NonOwnerClass}),
            itcenv, NONE)
    else raise TypeError("Declared class type: " ^ printTypeOpt(dType) ^ 
             " is not a valid class type.")

  | typeOODecl(InterfaceDeclIntern(Class{name, id, mode, superclass, interfaces,
        dType}), tenv, itcenv) =
    (*
     *  Typing an interface is the same as typing a class.
     *)
    (case (dType) of
        SOME t => raise TypeError("Interface " ^ name ^ " has type: " ^ 
                      printType(t) ^ " but interfaces can't have types.")
      | NONE => (empty_tenvironment, itcenv, NONE))

  | typeOODecl(MessageDeclIntern({name=msg_name,iType=evaledType}),
        tenv, itcenv) = 
    (*
     *  To type a message, we only need to add a new binding to the type env.
     *  Added (10/28/02): We also need to make sure that there is exactly 1
     *  marked class in a message.
     *)
    let
        val valid_fun_type = isValidFunType(evaledType)
        val owner_count = OwnerTypeCount(evaledType)
    in
        if (valid_fun_type) then
            if (owner_count = 1) then
                (new_tbinding(msg_name, evaledType), ({name=msg_name,
                    gblType=evaledType, pats=nil})::itcenv, NONE)
            else
                raise TypeError("Message " ^ msg_name ^ " has " ^ 
                    Int.toString(owner_count) ^ " owners; messages MUST " ^
                    "exactly one owner.")
        else raise TypeError("Message " ^ msg_name ^ " has type: " ^ 
             printType(evaledType) ^ " which is not a valid function type.")
    end

  | typeOODecl(MethodDeclIntern({name=msgName,formal=msgFormal,
        body=msgBody}), tenv, itcenv) = 
    (*
     *  To type a method, we need to make sure that the type of body matches
     *  the declared type of the function.
     *)
    let 
        val msgType = (lookup_tbinding(tenv, msgName) handle NotFound =>
            raise TypeError("Message name " ^ msgName ^ 
                            " not found in type env."))

        val (msgPatType, msgBodyType) = typeGetFunParts(msgType)
        val fml_tenv = typePattern(msgFormal, msgPatType)
        val extended_tenv = combine_tenvs(tenv, fml_tenv)
        val body_type = typeExpr(msgBody, extended_tenv)
        val firstAsSpecific = typeFirstAsSpecific(body_type, msgBodyType)

(*
 *      Add the method to our message list.
 *)
        val new_itcenv = itcenv_addElem(itcenv, nil, msgName, msgFormal)

    in
        if (firstAsSpecific) then
            (empty_tenvironment, new_itcenv,
             SOME({name=msgName, formal=msgFormal, body=msgBody}))
        else raise TypeError("Declared method type: " ^ printType(body_type) ^ 
            " does not match type of message with same name: " ^ 
            printType(msgBodyType))
    end

(*
 *  typeVDeclList() takes a list of decls an returns new type bindings.
 *)
and typeVDeclList([], tenv) = empty_tenvironment
  | typeVDeclList(first::rest, tenv) =
    (* 
     *  We first type check the first declaration.  We then type check the rest
     *  of the declarations, but with the type bindings from the first included.
     *)
    let 
        val first_tbindings:tEnvironment = typeVDecl(first, tenv) 
        val extended_tenv:tEnvironment = combine_tenvs(tenv, first_tbindings)
        val rest_tbindings:tEnvironment = typeVDeclList(rest, extended_tenv)
    in 
        combine_tenvs(first_tbindings, rest_tbindings) 
    end

(*
 *  typeOODeclList(): Takes a list of decls and run typeOODecl() on each decl.
 *      Returns the combination of all the decl's returns.
 *)
and typeOODeclList([], tenv, itcenv) = (empty_tenvironment, itcenv, nil)
  | typeOODeclList(first::rest, tenv, itcenv) =
    (* 
     *  We first type check the first declaration.  We then type check the rest
     *  of the declarations, but with the type bindings from the first included.
     *)
    let 
        val (first_tbindings, first_itcenv, first_methodenv) =
            typeOODecl(first, tenv, itcenv)
        val extended_tenv:tEnvironment = combine_tenvs(tenv, first_tbindings)
        val (rest_tbindings, new_itcenv, rest_methodenvs) =
            typeOODeclList(rest, extended_tenv, first_itcenv)
    in 
        (combine_tenvs(first_tbindings, rest_tbindings), new_itcenv,
            first_methodenv :: rest_methodenvs)
    end

(* 
 *  typeExpr returns the type of the given expression.
 *)
and typeExpr(VarExprIntern(var_name:string), tenv) =
    (*  
     *  To type check a var expr, we just need to get its type from the
     *  type environment.
     *)
    (lookup_tbinding(tenv, var_name) handle NotFound => 
        raise TypeError("Variable " ^ var_name ^ " not found in type env."))

(* 
 *  To type check a constant, just return the corresponding type.
 *)
  | typeExpr(IntExprIntern(v), tenv) = IntegerType
  | typeExpr(RealExprIntern(v), tenv) = RealType
  | typeExpr(StringExprIntern(v), tenv) = StringType
  | typeExpr(BoolExprIntern(v), tenv) = BooleanType
  | typeExpr(NilExprIntern(decType), tenv) = decType

(* 
 *  To type check a tuple expression, check each of the subexpressions, then 
 *  construct the tuple type with those values.
 *)
  | typeExpr(TupleExprIntern(exprs:ExprIntern list), tenv) =
    TupleType(typeExprList(exprs, tenv))

(*
 *  To type check a record, just return the type of that record.
 *)
  | typeExpr(RecordExprIntern(recs), tenv) = 
    let
(*
 *      This returns the type of the list of elements in the record.
 *)
        fun typeRecList(nil) = nil
          | typeRecList({eident=eID, evalue=eVal}::restRecs) =
                {tident=eID, ttype=typeExpr(eVal, tenv)}::typeRecList(restRecs)
    in
        RecordType(typeRecList(recs))
    end

(* 
 *  To type a unary operator expression, evaluate the argument, then type check
 *  the operator with the argument.
 *)
  | typeExpr(UnopExprIntern{operator:string, arg:ExprIntern}, tenv) =
    let 
        val arg_type = typeExpr(arg, tenv)
        val unop_type = typeUnop(operator, arg_type)
        val (unop_pat, unop_body) = typeGetFunParts(unop_type)
        val firstAsSpecific = typeFirstAsSpecific(arg_type, unop_pat)
    in 
        if (firstAsSpecific) then unop_body
        else raise TypeError("Input to unop does not match type of unop.")
    end

(* 
 *  To type a binary operator expression, type the arguments, then type the 
 *  operator with the arguments.
 *)
  | typeExpr(BinopExprIntern{arg1:ExprIntern, operator:string, 
                arg2:ExprIntern}, tenv) =
    let 
        val arg1_type = typeExpr(arg1, tenv)
        val arg2_type = typeExpr(arg2, tenv)
        val binop_type = typeBinop(operator, arg1_type)
        val (binop_pat, binop_body) = typeGetFunParts(binop_type)
        val (bop_pat_arg1, bop_pat_arg2) = typeGetBinopArgs(binop_pat)
        val arg1AsSpecific = typeFirstAsSpecific(arg1_type, bop_pat_arg1)
        val arg2AsSpecific = typeFirstAsSpecific(arg2_type, bop_pat_arg2)
    in
        if (arg1AsSpecific andalso arg2AsSpecific) then binop_body
        else raise TypeError("Arguments to binop do not match binop's type.")
    end

(*  
 *  To type check a call expr, we first type check the function and the 
 *  argument.  We then check to make sure that the type of the arg is the
 *  same as the input type to the call.  The returned type is the return
 *  type of the function.
 *)
  | typeExpr(CallExprIntern{function:ExprIntern, arg:ExprIntern}, tenv) =
    let 
        val call_type = typeExpr(function, tenv)
        val arg_type = typeExpr(arg, tenv)
    in
        case (call_type) of
            FunctionType(pat_type, body_type) => 
            let
                val firstAsSpecific = typeFirstAsSpecific(arg_type, pat_type)
            in
                if (firstAsSpecific) then body_type
                else raise TypeError("Argument type: " ^ printType(arg_type) ^ 
                        " to call doesn't match input type: " ^ 
                        printType(pat_type) ^ ".")
            end
          
          | _ => raise TypeError("Call expr expected a function or class call," ^
                     " but encountered neither.")
    end

(*  
 *  To type an if expr, we need to make sure that the type of the if case is 
 *  boolean, and that the two cases have the same type.
 *)		  
  | typeExpr(IfExprIntern{testExpr:ExprIntern, thenExpr:ExprIntern, 
                elseExpr:ExprIntern}, tenv) =
    let 
        val test_type = typeExpr(testExpr, tenv)
        val then_type = typeExpr(thenExpr, tenv)
        val else_type = typeExpr(elseExpr, tenv)
        val thenAsSpecific = typeFirstAsSpecific(then_type, else_type)
        val elseAsSpecific = typeFirstAsSpecific(else_type, then_type)
    in
        case (test_type) of
            BooleanType => (if (thenAsSpecific andalso elseAsSpecific) then 
                            then_type else 
                            raise TypeError("Then type and Else type don't " ^
                                            "match."))
          | _ => raise TypeError("Test type must be a Boolean type.")
    end

(*
 *  To type check a let expr, first type check the decl list, getting back
 *  the new type bindings.  Then use an extended type environement to type
 *  check the body expr.
 *)
  | typeExpr(LetExprIntern{decls:vDeclIntern list, body:ExprIntern}, tenv) =
    let
        val tbindings:tEnvironment = typeVDeclList(decls, tenv)
        val extended_tenv:tEnvironment = combine_tenvs(tenv, tbindings)
    in 
        typeExpr(body, extended_tenv)
    end

  | typeExpr(ClassExprIntern(classInst as Class{name, id, mode, superclass, 
             interfaces, dType=SOME(clType)}), tenv) = 
        (*
         * COLIN: Again, just assigning non owner type...
         *)
        FunctionType(clType, ClassInstanceType{class=classInst,
            is_owner=NonOwnerClass})

  | typeExpr(errExpr, tenv) = 
        raise TypeError("No type checking algorithm found for " ^ printExprIntern(errExpr))

and typeExprList(nil, tenv) = nil
  | typeExprList(head_expr::rest_exprs, tenv) =
    let
        val head_type = typeExpr(head_expr, tenv)
        val rest_types = typeExprList(rest_exprs, tenv)
    in
        head_type::rest_types
    end

(*
 *  typePattern() type checks a pattern, returning the type of the pattern.
 *  
 *  It is given a pattern tcomp that must agree with typewise.  If there is a
 *  disagreement, a TypeError exception will be raised.
 *)
and typePattern(VarPatternIntern(name), tcomp) =
    let
        val new_tenv = new_tbinding(name, tcomp)
    in
        new_tenv
    end

  | typePattern(WildcardPatternIntern, tcomp) =
    let
        val new_tenv = new_tbinding("_", tcomp)
    in
        new_tenv
    end

  | typePattern(IntPatternIntern(_), tcomp) = 
    (case (tcomp) of
        IntegerType => empty_tenvironment
      | _ => raise TypeError("Pattern type (int) doesn't match input type."))

  | typePattern(RealPatternIntern(_), tcomp) = 
    (case (tcomp) of
        RealType => empty_tenvironment
      | _ => raise TypeError("Pattern type (real) doesn't match input type."))

  | typePattern(StringPatternIntern(_), tcomp) = 
    (case (tcomp) of
        StringType => empty_tenvironment
      | _ => raise TypeError("Pattern type (string) doesn't match input type."))

  | typePattern(BoolPatternIntern(_), tcomp) =
    (case (tcomp) of
        BooleanType => empty_tenvironment
      | _ => raise TypeError("Pattern type (string) doesn't match input type."))

  | typePattern(NilPatternIntern, tcomp) = empty_tenvironment

  | typePattern(TuplePatternIntern(patList), tcomp) = 
    let
        val tcomp_list = typeGetTypesFromTuple(tcomp)
        val list_tenv = typePatterns(patList, tcomp_list)
    in
        list_tenv
    end

  | typePattern(RecordPatternIntern(pRecList), RecordType(tRecList)) =
    let
        fun compareRecLists(nil, nil) = empty_tenvironment
          | compareRecLists(nil, notNil) = raise TypeError("Record's type: " ^
                printType(RecordType(tRecList)) ^ " has more elements than its " ^
                " pattern: " ^ printPatternIntern(RecordPatternIntern(pRecList)))

          | compareRecLists(notNil, nil) = raise TypeError("Record's type: " ^
                printType(RecordType(tRecList)) ^ " has less elements than its " ^
                " pattern: " ^ printPatternIntern(RecordPatternIntern(pRecList)))

          | compareRecLists({ident, pattern}::pList, {tident, ttype}::tList) =
            if (ident = tident) then combine_tenvs(typePattern(pattern, ttype),
                                        compareRecLists(pList, tList))
            else
                raise TypeError("Record id: " ^ ident ^ " doesn't match type " ^
                    " id: " ^ tident)
    in
        compareRecLists(pRecList, tRecList)
    end

  | typePattern(ConsPatternIntern{head, tail}, tcomp) =
    let
        val tcomp_head = typeGetTypeFromList(tcomp)
        val head_tenv = typePattern(head, tcomp_head)
        val tail_tenv = typePattern(tail, tcomp)
        val return_tenv = combine_tenvs(head_tenv, tail_tenv)
    in
        return_tenv
    end

(*
 * COLIN: more confusion over non owner type stuff...
 *)
  | typePattern(ClassPatternIntern({class=cl1 as Class{name, id, mode,
        superclass, interfaces, dType=SOME(clType)}, rep}),
        ClassInstanceType{class=cl2 as Class{name=sn, id=sid, mode=sm,
            superclass=scl, interfaces=sicl, dType=SOME(superType)},
            is_owner=owner}) =
    if (FirstIsSubclassOfSecond(cl1, cl2)) then
        let
            val _ = print("DEBUG: class typing: " ^ printPatternIntern(rep) ^
                " against " ^ printType(clType) ^ "\n")
            val rep_tenv = typeClassPattern(rep, clType)
        in
            rep_tenv
        end
    else raise TypeError("Class: " ^ name ^ " isn't a subclass of: " ^ sn)

(*
 *  If the case above didn't catch the class pattern than we have an interface
 *  as our comparing type.
 *)
(*
 * COLIN: still not sure what to do with owner info.
 *)
  | typePattern(ClassPatternIntern({class=cl1 as Class{name, id, mode,
        superclass, interfaces, dType=SOME(clType)}, rep}),
        ClassInstanceType{class=interface, is_owner=owner}) =
    if (classImplementsInterface(cl1, interface)) then
        let
            val _ = print("DEBUG: class typing: " ^ printPatternIntern(rep) ^
                " against " ^ printType(clType))
            val rep_tenv = typeClassPattern(rep, clType)
        in
            rep_tenv
        end
    else raise TypeError("Class: " ^ name ^ " does not implement: " ^
        printClassName(interface))

  | typePattern(ClassPatternIntern({class=Class{name, id, mode, superclass,
                interfaces, dType=NONE}, rep}), tcomp) =
    raise TypeError("We don't type Interfaces; this shouldn't have been called!")
  
  | typePattern(otherPat, tcomp) = raise TypeError("I don't know how to" ^
        " type check this pattern: " ^ printPatternIntern(otherPat) ^ " with "
        ^ "given type of: " ^ printType(tcomp))

and typePatterns(nil, tcomp) = empty_tenvironment
  | typePatterns(pat::pats, tcomp) = 
    let
        fun getTcompParts(tcomp_head::tcomp_rest) = (tcomp_head, tcomp_rest)
          | getTcompParts(_) = raise TypeError("Incoming pattern must be a list!")

        val (tcomp_h, tcomp_r) = getTcompParts(tcomp)
        val pat_tenv = typePattern(pat, tcomp_h)
        val pats_tenv = typePatterns(pats, tcomp_r)
        val new_tenv = combine_tenvs(pat_tenv, pats_tenv)
    in
        new_tenv
    end
    
(*
 *  To type a class pattern, we know that the record and type must both be
 *  records.  From there, we check that the elements in the record's pattern
 *  are a subset of the records in the record's type.  If so, then return the
 *  type of the pattern's record list.
 *)
and typeClassPattern(RecordPatternIntern(recPat), RecordType(recType)) = 
    let
(*
 *      typeOneElt() takes a record pattern element and the record's type list
 *      and finds the corresponding element in that list.  When the element is
 *      found, it type checks the pattern of the element with the found type.
 *      If no match is found, and error is raised.
 *)
        fun typeOneElt(elt as {ident=id, pattern=pat}, nil) = 
(*
 *          The list of types is empty, and no match was found for the name, so
 *          an error is raised.
 *)   
            raise TypeError("Found no match for: " ^ id ^ " in record's type.")
          | typeOneElt(elt as {ident=pID, pattern=pPat},
                {tident=tID, ttype=tType}::rest) =
            if (pID = tID) then
(*
 *              The names match, so we type check the pattern, return the env
 *              created from the type check.
 *)
                typePattern(pPat, tType)
            else
(*
 *              The names don't match, so check the rest of the list.
 *)
                typeOneElt(elt, rest)

(*
 *      typeAllElts() takes the list of the record's pattern elements and checks
 *      each one against the type list, using typeOneElt(), and returns the 
 *      catenation of all the type environments created from the typeOneElt()
 *      calls.
 *)
        fun typeAllElts(nil, typeList) = empty_tenvironment
          | typeAllElts(first::rest, typeList) = 
                combine_tenvs(typeOneElt(first, typeList),
                    typeAllElts(rest, typeList))
    in
        typeAllElts(recPat, recType)
    end

(*
 *  We can also return valid if the incoming pattern is a WildcardPattern.
 *)
  | typeClassPattern(WildcardPatternIntern, RecordType(recType)) =
        typePattern(WildcardPatternIntern, RecordType(recType))

(*
 *  We can also return valid if the incoming pattern is a variable pattern.
 *)
  | typeClassPattern(VarPatternIntern(vname), RecordType(recType)) =
        typePattern(VarPatternIntern(vname), RecordType(recType))

  | typeClassPattern(RecordPatternIntern(_), _) =
        raise TypeError("Class has invalid type.")
  | typeClassPattern(_, RecordType(_)) =
        raise TypeError("Class has invalid pattern.")
  | typeClassPattern(_, _) =
        raise TypeError("Class has invalid pattern and type.")

(* 
 *  typeUnop() takes a sting (the operator) and the parameter type to return
 *  the input and return types (as a FunctionType) needed for operator to
 *  evaluate corretly.
 *)
and typeUnop("~", IntegerType) = FunctionType(IntegerType, IntegerType)
  | typeUnop("~", RealType) = FunctionType(RealType, RealType)
  | typeUnop("not", BooleanType) = FunctionType(BooleanType, BooleanType)
  | typeUnop(unop, t) = raise TypeError("Unknown unop and type combination (" ^
        unop ^ "; " ^ printType(t) ^ ")")

(*
 *  typeBinop() does the same thing as typeUnop(), only for binary operations.
 *  
 *  <colin> 1/11/02 We currently supporting all the variations that we should.
 *  Any changes made here should be changed in emlEval.sml too!
 *)
and typeBinop("+", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), IntegerType)
  | typeBinop("+", RealType) = FunctionType(TupleType([RealType, RealType]),
        RealType)
  | typeBinop("-", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), IntegerType)
  | typeBinop("-", RealType) = FunctionType(TupleType([RealType, RealType]),
        RealType)
  | typeBinop("*", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), IntegerType)
  | typeBinop("*", RealType) = FunctionType(TupleType([RealType, RealType]),
        RealType)
  | typeBinop("div", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), IntegerType)
  | typeBinop("/", RealType) = FunctionType(TupleType([RealType, RealType]),
        RealType)
  | typeBinop("mod", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), IntegerType)
  | typeBinop("<", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop("<", RealType) = FunctionType(TupleType([RealType, RealType]),
        BooleanType)
  | typeBinop("<", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop("<=", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop("<=", RealType) = FunctionType(TupleType([RealType, RealType]),
        BooleanType)
  | typeBinop("<=", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop(">=", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop(">=", RealType) = FunctionType(TupleType([RealType, RealType]),
        BooleanType)
  | typeBinop(">=", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop(">", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop(">", RealType) = FunctionType(TupleType([RealType, RealType]),
        BooleanType)
  | typeBinop(">", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop("^", StringType) = FunctionType(TupleType([StringType,
        StringType]), StringType)
  | typeBinop("::", IntegerType) = FunctionType(TupleType([IntegerType,
        ListType(IntegerType)]), ListType(IntegerType))
  | typeBinop("::", RealType) = FunctionType(TupleType([RealType,
        ListType(RealType)]), ListType(RealType))
  | typeBinop("::", StringType) = FunctionType(TupleType([StringType,
        ListType(StringType)]), ListType(StringType))
  | typeBinop("::", BooleanType) = FunctionType(TupleType([BooleanType,
        ListType(BooleanType)]), ListType(BooleanType))
  | typeBinop("=", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop("=", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop("=", BooleanType) = FunctionType(TupleType([BooleanType,
        BooleanType]), BooleanType)
  | typeBinop("<>", IntegerType) = FunctionType(TupleType([IntegerType,
        IntegerType]), BooleanType)
  | typeBinop("<>", StringType) = FunctionType(TupleType([StringType,
        StringType]), BooleanType)
  | typeBinop("<>", BooleanType) = FunctionType(TupleType([BooleanType,
        BooleanType]), BooleanType)
  | typeBinop(binop, t) = raise TypeError("Unknown binop and type combination ("
        ^ binop ^ ", " ^ printType(t) ^ ")")

(*
 *  typeFirstAsSpecific() takes two types and returns true if the first type is
 *  at least as specific as the second type.  Otherwise it returns false.
 *)
and typeFirstAsSpecific(IntegerType, IntegerType) = true
  | typeFirstAsSpecific(RealType, RealType) = true
  | typeFirstAsSpecific(BooleanType, BooleanType) = true
  | typeFirstAsSpecific(StringType, StringType) = true
  | typeFirstAsSpecific(TupleType(types1), TupleType(types2)) =
        if (length(types1) <> length(types2)) then false
        else typeAllFirstAsSpecific(types1, types2)
  | typeFirstAsSpecific(RecordType(recs1), RecordType(recs2)) = 
        if (length(recs1) <> length(recs2)) then false
        else typeRecsAsSpecific(recs1, recs2)

  | typeFirstAsSpecific(ListType(type1), ListType(type2)) = 
        typeFirstAsSpecific(type1, type2)

(*
 * More owner stuff, probably don't care about that here.
 *)
  | typeFirstAsSpecific(ClassInstanceType{class=class1, is_owner=own1},
                        ClassInstanceType{class=class2, is_owner=own2}) =
        FirstIsSubtypeOfSecond(class1, class2)

  | typeFirstAsSpecific(FunctionType(f1_arg, f1_ret),
                        FunctionType(f2_arg, f2_ret)) =
    let
        val fun_args_comp = typeFirstAsSpecific(f1_arg, f2_arg)
        val fun_ret_comp = typeFirstAsSpecific(f1_ret, f2_ret)
    in
        if (fun_args_comp andalso fun_ret_comp) then true
        else false
    end
  
  | typeFirstAsSpecific(_, _) = false

and typeAllFirstAsSpecific(nil, nil) = true
  | typeAllFirstAsSpecific(nil, _) = false
  | typeAllFirstAsSpecific(_, nil) = false
  | typeAllFirstAsSpecific(t1_head::t1_rest, t2_head::t2_rest) = 
    let 
        val head_comp = typeFirstAsSpecific(t1_head, t2_head)
        val rest_comp = typeAllFirstAsSpecific(t1_rest, t2_rest)
    in
        if (rest_comp) then head_comp
        else false 
   end

and typeRecsAsSpecific(nil, nil) = true
  | typeRecsAsSpecific(nil, notNil) = false
  | typeRecsAsSpecific(notNil, nil) = false
  | typeRecsAsSpecific({tident=tID1, ttype=type1}::r1Rest,
        {tident=tID2,ttype=type2}::r2Rest) = 
    if (typeFirstAsSpecific(type1, type2)) then 
        typeRecsAsSpecific(r1Rest, r2Rest)
    else false

(*
 *  itc():  Implementation side type checking.
 *
 *  Parameters: a type environment [<colin> not needed?] and an itc environment
 *      (a list of all messages, represented by a name, type and pattern list)
 *
 *  Return value: if the message list is neither incomplete nor ambiguous, this
 *      returns true.  Otherwise and exception (TypeError) is raised, which
 *      should be handled by the repl loop.
 *
 *  Comments: given the list of messages, itc() first computes whether the
 *      first message satisfies the completeness and ambiguity problems.  If
 *      it does, it recursively calls itc() on the rest of the message list.
 *      When the message list is exhausted, true is returned.
 *)
and itc(_, nil) = true
  | itc(name_env, (msg as {name=mname, gblType=mtype, pats=mpats})::msgList) =
    let
(*
 *      We don't really care about these return values.  If either of these
 *      tests fail, they will raise an exception, which will bump execution
 *      back up to the top level.
 *)
        val _ = itc_completeness(mtype, mpats, name_env)
        val _ = itc_ambiguity(mtype, mpats, mpats)
    in
        itc(name_env, msgList)
    end

(*
 *  itc_ambiguity(): Check to insure a message has no ambiguities.
 *
 *  Parameters: the type of the message, the methods left to check, and the
 *      complete list of methods in the message.
 *
 *  Return value: if the message is not ambiguous, true is returned.  If it
 *      is ambiguous, then an exception (TypeError) is raised, citing the
 *      offending patterns.
 *
 *  Comments: this loops recursively on the second parameter.  Given a list
 *      of methods, this performs a pairwise check on each pair of methods.
 *      This is done by taking a single method and the remaining methods after
 *      it in the list.  We then compare the selected method to every method
 *      in the remainder of the list.  If there are no ambiguities, we
 *      recursively call ourselves with the rest of the list.  This insures
 *      that every pair is checked.  We need an additional paremter (the
 *      complete list of methods in a message) because we may need a third
 *      method to insure there is no ambiguity.
 *)
and itc_ambiguity(mtype, nil, pats) = true
  | itc_ambiguity(mtype, hdPat::tlPats, pats) =
    let
(*
 *      ambiguity_pairs(): Compares a pattern to a list of patterns.
 *
 *      Parameters: the pattern being checked and the list of patterns to check
 *          the pattern against.
 *
 *      Return value: returns true if the pattern is not ambiguous with any
 *          patterns in the list, raises TypeError if it is not.
 *
 *      Comments: helper function for itc_ambiguity().  Compares every pair
 *          using ambiguity_pair().
 *)
        fun ambiguity_pairs(basePat, nil) = true
          | ambiguity_pairs(basePat, compPat::compPats) =
            let
                val _ = ambiguity_pair(basePat, compPat, pats)
            in
                ambiguity_pairs(basePat, compPats)
            end

(*
 *      This declaration insures that the current pattern is not ambiguous.
 *)
        val _ = ambiguity_pairs(hdPat, tlPats)
    in
(*
 *      If we make it this far, then we know that the current pattern is good,
 *      so we check the rest of the patterns.
 *)
        itc_ambiguity(mtype, tlPats, pats)
    end

(*
 *  ambiguity_pair(): compares two patterns to see if they are ambiguous.
 *
 *  Parameters: two patterns and a list of all other patterns in the same
 *      message.
 *
 *  Return value: if the pair is not ambiguous, then true is returned.  If they
 *      are ambiguous, an exception (TypeError) is raised.
 *
 *  Comments: we need the third parameter because it is used to search for a
 *      third (more specific) pattern if we find some intersection of these
 *      two.
 *)
and ambiguity_pair(pat1, pat2, otherPats) =
    let
        val patSpecificity = compareSpecificity(pat1, pat2)
    in
(*
 *      If either pattern is more specific than the other, we can stop and
 *      signal that they aren't ambiguous.
 *)
        case (patSpecificity) of
            FirstIsMoreSpecific => true
          | SecondIsMoreSpecific => true
          | _ =>
(*
 *          If neither is more specific, we check to see if the two patterns
 *          have an intersection.  If they don't then there does not exist a
 *          case that would be ambiguous between this pair.
 *)
            let
                val intersectPat = patternIntersect(pat1, pat2)
            in
                case (intersectPat) of
                    NONE => true
                  | SOME(iPat) =>
(*
 *                  If they do have an intersection, we search the list of
 *                  methods looking for a pattern that matches the intersect
 *                  and is also more specific than both matches.  If we find
 *                  one we are okay.  If we don't, then the current pair is
 *                  ambiguous.
 *)
                    let
                        val isBetterPat = findBetterPat(otherPats, iPat,
                            pat1, pat2)
                    in
                        if (isBetterPat = true) then true
                        else raise TypeError("Pattern (" ^
                                printPatternIntern(pat1) ^ ") and pattern (" ^
                                printPatternIntern(pat2) ^ ") are ambiguous.")
                    end
            end
    end

(*
 *  patternIntersect(): Takes two patterns and creats a pattern that is the
 *      union of those two patterns.
 *
 *  Parameters: two internalized patterns
 *
 *  Return value: a internalized pattern option.
 *
 *  Comments: If no intersect can be computed, then NONE is returned.
 *  Otherwise SOME with the internalized pattern is returned.
 *)
and patternIntersect(VarPatternIntern(_), pat2) = SOME(pat2)
  | patternIntersect(pat1, VarPatternIntern(_)) = SOME(pat1)
  | patternIntersect(WildcardPatternIntern, pat2) = SOME(pat2)
  | patternIntersect(pat1, WildcardPatternIntern) = SOME(pat1)
  | patternIntersect(IntPatternIntern(i1), IntPatternIntern(i2)) =
        if (i1 = i2) then SOME(IntPatternIntern(i1)) else NONE
  | patternIntersect(RealPatternIntern(r1), RealPatternIntern(r2)) =
        if (r1 <= r2 andalso r1 >= r2) then SOME(RealPatternIntern(r1))
        else NONE
  | patternIntersect(StringPatternIntern(s1), StringPatternIntern(s2)) =
        if (s1 = s2) then SOME(StringPatternIntern(s1)) else NONE
  | patternIntersect(BoolPatternIntern(b1), BoolPatternIntern(b2)) =
        if (b1 = b2) then SOME(BoolPatternIntern(b1)) else NONE
  | patternIntersect(NilPatternIntern, NilPatternIntern) =
        SOME(NilPatternIntern)
  | patternIntersect(TuplePatternIntern(p1s), TuplePatternIntern(p2s)) =
        patternIntersectList(p1s, p2s, nil)
  | patternIntersect(ConsPatternIntern{head=h1, tail=t1},
        ConsPatternIntern{head=h2, tail=t2}) = 
    let
        val hdIntersect = patternIntersect(h1, h2)
        val tlIntersect = patternIntersect(t1, t2)
    in
        case (hdIntersect) of
            NONE => NONE
          | SOME(hdt) => case (tlIntersect) of
                NONE => NONE
              | SOME(tlt) => SOME(ConsPatternIntern{head=hdt, tail=tlt})
    end
  | patternIntersect(RecordPatternIntern(recs1), RecordPatternIntern(recs2)) =
        patternIntersectRecs(recs1, recs2, nil)
  | patternIntersect(ClassPatternIntern{class=c1, rep=r1},
        ClassPatternIntern{class=c2, rep=r2}) =
    if (FirstIsSubclassOfSecond(c1, c2)) then
        let
            val intersectPat = patternIntersect(r1, r2)
        in
            case (intersectPat) of
                NONE => NONE
              | SOME(p) => SOME(ClassPatternIntern{class=c1, rep=p})
        end
    else if (FirstIsSubclassOfSecond(c2, c1)) then
        let
            val intersectPat = patternIntersect(r2, r1)
        in
            case (intersectPat) of
                NONE => NONE
              | SOME(p) => SOME(ClassPatternIntern{class=c2, rep=p})
        end
    else NONE
  | patternIntersect(_, _) = NONE

(*
 *  patternIntersectRecs(): create an intersecting pattern for two records.
 *
 *  Parameters: two lists from RecordPatternIntern() patterns and base off
 *      which to create a new list (should be started as nil).
 *
 *  Return value: returns an RecordPatternIntern option.  That is, if there is
 *      a record that intersects both parameters, then
 *      SOME(RecordPatternIntern) is returned.  Otherwise, NONE is returned.
 *
 *  Comments: handles records assuming that the first record might be a
 *      subclass of the second (but the second cannot be a subclass of the
 *      first).
 *)
(*
 *  If we have no lists but some list, we create a new record.
 *)
and patternIntersectRecs(nil, nil, newRecs) =
        SOME(RecordPatternIntern(newRecs))
(*
 *  If the second list is empty, but the first is not, we assum that the first
 *  is a subclass of the second, so we create a new record with our new list
 *  plus the remaining elements from the subclass.
 *)
  | patternIntersectRecs(superRecs, nil, newRecs) =
        SOME(RecordPatternIntern(newRecs@superRecs))
(*
 *  If the first list is empty but the second is not, we assume some bizzare
 *  error has occured and we bail.
 *)
  | patternIntersectRecs(nil, notNil, _) = raise TypeError("Record intersect "
        ^ "recieved a second record with alphabetical precednce.  I didn't " ^
        "think this would happen!")
  | patternIntersectRecs((srec as {ident=sid, pattern=spat})::srecs,
        (crec as {ident=cid, pattern=cpat})::crecs, newRecs) =
(*
 *  If the idents are equal, then the intersect is a part of the total
 *  intersect.
 *)
    if (sid = cid) then
        let
            val curIntersect = patternIntersect(spat, cpat)
        in
            case (curIntersect) of
(*
 *              If this pattern has no intersection, than the whole record has
 *              no intersection.  If it does have an intersection, then we add
 *              that intersection to our new record list and call again on the
 *              rest of the record.
 *)
                NONE => NONE
              | SOME(p) => patternIntersectRecs(srecs, crecs,
                    newRecs @ [{ident=sid, pattern=p}])
        end

(*
 *      If the first id precedes the second one (alphabetically) then we assume
 *      that we are dealing with a subclass and we just add the first record to
 *      our list.
 *)
        else if (sid < cid) then
            patternIntersectRecs(srecs, crec::crecs, newRecs @ [srec])
(*
 *      Otherwise the second pattern precedes the first.  This should never
 *      happen.
 *)
        else raise TypeError("Record intersect recieved a second record with "
            ^ "alphabetical precednce.  I didn't think this would happen!")

(*
 *  patternIntersectList(): computes the intersect of a list of patterns
 *
 *  Parameters: the two lists to intersect with and a base list to create the
 *      new list with (should be nil!)
 *
 *  Return value: returns a TuplePatternIntern option containing the
 *      intersection of each element in the lists.
 *
 *  Comments: if there is an empty intersection at any element, then the whole
 *      list is an empty intersection.
 *)
(*
 *  The lists are emtpy but we have some patterns, so we create a new pattern.
 *)
and patternIntersectList(nil, nil, newPats) = SOME(TuplePatternIntern(newPats))
(*
 *  The second list has more elements, the two patterns aren't comperable (I
 *  don't think this will ever occur).
 *)
  | patternIntersectList(nil, notNil, _) = NONE
(*
 *  The first list has more elements, the two patterns aren't comperable (I
 *  don't think this will ever occur).
 *)
  | patternIntersectList(notNil, nil, _) = NONE
  | patternIntersectList(pat1::pats1, pat2::pats2, newPats) =
    let
(*
 *      Compute the intersection of the current element.
 *)
        val hdPatIntersect = patternIntersect(pat1, pat2)
    in
(*
 *      If there is no intersection for this element, then there is no
 *      intersection at all.  If there is an intersection, then we append the
 *      intersection to our new list and intersect the rest of the list.
 *)
        case (hdPatIntersect) of
            NONE => NONE
          | SOME(iPat) =>
                patternIntersectList(pats1, pats2, newPats @ [iPat])
    end

(*
 *  findBetterPat(): searches a list for a best fit pattern.
 *
 *  Parameters: the list of patterns to search, a pattern that must be more
 *      specific than the sought for pattern, and two patterns that the sought
 *      for pattern must be more specific than.
 *
 *  Return value: if a better pattern is found, return true (the identity of
 *      better pattern is not important).  If a pattern can not be found,
 *      return false.
 *
 *  Comments: given a list of patterns, we search of a pattern can be used if
 *      the intersectPat is ever created.  That is, both match1 and match2 can
 *      handle the intersectPat, but neither is more specific than the other.
 *      We are searching for a pattern that can also handle the intersect and
 *      is more specific than either of the original matches.
 *)
(*
 *  If the list is empty, we've exhausted all options, no better pattern found.
 *)
and findBetterPat(nil, _, _, _) = false
  | findBetterPat(hdPat::tlPats, intersectPat, match1, match2) =
    let
(*
 *      See if the new pattern can even handle the intersect pattern.
 *)
        val fitsIntersect = FirstIsAsSpecificAsSecond(intersectPat, hdPat)
    in
(*
 *      If the new pattern does handle the intersect, check to see if it is
 *      strictly more specific than both other matches.  If it is not, then
 *      continue searching the list.  If the new pattern does not handle the
 *      intersect, then also continue searching the list.
 *)
        if (fitsIntersect = true) then
        let
            val beatsM1 = compareSpecificity(hdPat, match1)
            val beatsM2 = compareSpecificity(hdPat, match2)
        in
            if (beatsM1 = FirstIsMoreSpecific andalso
                beatsM2 = FirstIsMoreSpecific) then true
            else findBetterPat(tlPats, intersectPat, match1, match2)
        end
        else findBetterPat(tlPats, intersectPat, match1, match2)
    end

(*
 *  itc_completeness(): Check to make sure than a message is complete.
 *
 *  Parameters: The type of message, the list of patterns of the message, and
 *      the name environment.
 *
 *  Return value: If the message is complete, true is returned.  If it is not,
 *      an error is raised citing which generated pattern is not handled (if
 *      more than one, the first unhandled pattern found is cited).
 *
 *  Comments: itc_completeness() first generates a list of all possible cases
 *      that could be created matching the message (using genPats()).  It then
 *      takes the list of generated patterns and checks to see if there is a
 *      a pattern in the method list that can handle it.  If all generated
 *      patterns can be handled, then this message is complete.  If a single
 *      generated pattern can not be handled, an exception is raised.
 *)
and itc_completeness(mtype, mpats, name_env) =
    let
        val generatedPats:PatternIntern list =
                genPats(mpats, mtype, name_env)
        val _ = print("DEBUG: Possible patterns:\n" ^
                printPatternInternList(generatedPats) ^ "\n")

(*
 *      checkGeneratedPats(): Takes a list of patterns and checks them for
 *          completeness.
 *
 *      Parameters: a list of patterns to check.
 *
 *      Return value: If all patterns are covered, true is returned.  If not,
 *          an exception is raised.
 *
 *      Comments: We only need to pass the list of patterns to check (and not
 *          the list of patterns to check with) because we can use the mpats
 *          variable (since we are local to the itc_completeness() function).
 *)
        fun checkGeneratedPats(nil) = true
          | checkGeneratedPats(gPat::gPats) =
            let
(*
 *              checkGenPat(): Takes a list of potential matches to a specific
 *                  pattern, checks them to see if any potentials cover it.
 *
 *              Parameters: The list of potential matches.
 *
 *              Return value: If a match is found, true is returned.  Otherwise
 *                  an exception is raised.
 *
 *              Comments: We need the list of patterns to check with because we
 *                  recurse through them.  We don't need the pattern to match
 *                  against because it is declared outside this function.
 *)
                fun checkGenPat(nil) = raise TypeError("Incomplete!  Potential"
                        ^ " pattern " ^ printPatternIntern(gPat) ^ " has no " ^
                        "case to handle it.")
                  | checkGenPat(potentialPat::rest) =
(*
 *                  If the generated pattern is at least as specific as our
 *                  potential match, then we know that our match will cover it.
 *)
                    if (FirstIsAsSpecificAsSecond(gPat, potentialPat)) then
                        true
                    else checkGenPat(rest)

(*
 *              Use this to get the function running.  We don't care about the
 *              return value -- an exception will be raised if this doesn't
 *              pass.
 *)
                val _ = checkGenPat(mpats)
            in
(*
 *              If we make it here, then the current generated pattern has
 *              found a match.  Continue with the rest of the patterns.
 *)
                checkGeneratedPats(gPats)
            end
(*
 *      Use this to get the function running.  We don't care about the return
 *      value -- an exception will be raised if this doesn't pass.
 *)
        val _ = checkGeneratedPats(generatedPats)
    in
(*
 *      If all cases have been handled (determined by checkGeneratedPats()),
 *      then this message is complete!
 *)
        true
    end

(*
 *  genPats(): computes a list of patterns that fit a specified type
 *
 *  Parameters: The patterns to generate cases for, the type that the generated
 *      patterns must fit, and a name environment.
 *
 *  Return value: A list of patterns that could possible be generated.
 *
 *  Comments: Use the pattern list to determine the depth we must go to in
 *      our pattern generation.  We need the name environment for calculating
 *      the class patterns.
 *)
and genPats(mpats, mtype, name_env):PatternIntern list =
    let
(*
 *      Since the type is a function type, we only want the first part of the
 *      type.
 *)
        val (argType, _) = typeGetFunParts(mtype)

(*
 *      Determine the depth our patterns must be generated to.  If there are no
 *      patterns, we instead get the depth of the type.
 *)
        val patDepth = if (length(mpats) > 0) then depthPatterns(mpats, 0)
                       else  depthType(argType)



        val _ = print("DEBUG: Depth of pattern is: " ^ Int.toString(patDepth) ^ "\n")
        val _ = print("DEBUG: Type of pattern is: "^ printType(argType) ^ "\n")

(*
 *      generateDepthPats(): returns a list of patterns that match a type given
 *          a specific depth.
 *
 *      Parameters: the depth that we need to generate patterns to
 *          (specificity).
 *
 *      Return value: A list of patterns that fit the type given.
 *
 *      Comments: <colin> not yet finished (lists aren't supported). Also
 *          unsure of correctness of class, tuple and record generation.
 *)
        fun generateDepthPats(0, _):PatternIntern list = [WildcardPatternIntern]
          | generateDepthPats(d, IntegerType) = [WildcardPatternIntern]
          | generateDepthPats(d, RealType) = [WildcardPatternIntern]
          | generateDepthPats(d, BooleanType) = BoolPatternIntern(true) ::
                BoolPatternIntern(false) :: nil
          | generateDepthPats(d, StringType) = [WildcardPatternIntern]
          | generateDepthPats(d, FunctionType(_, _)) = [WildcardPatternIntern]
          | generateDepthPats(d, ListType(lType)) =
            let
(*
 *              Treat lists as an abstract class with concrete subclasses nil
 *              and cons{hd:T, lt:T List}.  The nil class can only be
 *              represented by the nil pattern.  The cons class is handled by
 *              computing all the patterns that can match type lType (the head)
 *              and all the patterns that match type lType (the tail) (with a
 *              depth of d-1 to insure an end to the recursion).  The resulting
 *              patterns are then cross producted together to create a list of
 *              cons patterns that match the given type.
 *)

(*
 *              Head pats is a list of all patterns that match the head part of
 *              our cons pattern.  Tail pats is a list of all patterns that
 *              match the tail part of our cons pattern.
 *)
                val headPats:PatternIntern list = generateDepthPats(d-1, lType)
                val tailPats:PatternIntern list =
                        generateDepthPats(d-1, ListType(lType))

(*
 *              Given a single head pattern and a list of tail pages, this
 *              function creates a list of ConsPatternIntern containing the
 *              head and each tail pattern.
 *)
                fun combineHeadAndTails(hPat:PatternIntern,
                        nil:PatternIntern list) = nil:PatternIntern list
                  | combineHeadAndTails(hPat, tPat::tPats) =
                    let
                        val restCombinedPats:PatternIntern list =
                                combineHeadAndTails(hPat, tPats)
                        val curPat = ConsPatternIntern{head=hPat, tail=tPat}
                    in
                        curPat :: restCombinedPats
                    end

(*
 *              Given a list of head patterns and a list of tail patterns, this
 *              function takes each head pattern and combines it with all the
 *              tail patterns (using combineHeadAndTails()).  It does this for
 *              all the head patterns and combines the resulting cons patterns
 *              into the return list.
 *)
                fun combineConsHeadsAndTails(nil:PatternIntern list,
                        tails:PatternIntern list) = nil:PatternIntern list
                  | combineConsHeadsAndTails(headPat::restPats, tailPats) =
                    let
                        val restConsPats:PatternIntern list  =
                                combineConsHeadsAndTails(restPats, tailPats)
                        val curConsPats:PatternIntern list =
                                combineHeadAndTails(headPat, tailPats)
                    in
                        curConsPats @ restConsPats
                    end

(*
 *              Generate our patterns.
 *)
                val NilPat = NilPatternIntern
                val ConsPats:PatternIntern list = combineConsHeadsAndTails(headPats,
                        tailPats)
            in
                NilPat :: ConsPats
            end
          | generateDepthPats(d, TupleType(tList)):PatternIntern list =
            let
(*
 *              genTuplePats(): for each pattern in our pattern list, generate
 *                  the pattern list for that element.
 *
 *              Parameters: the list of patterns to generate through
 *
 *              Return value: a list of pattern match sets (a list of lists)
 *
 *              Comments: 
 *)
                fun genTuplePats(nil) = nil
                  | genTuplePats(pat::pats) =
                    let
                        val curPats = generateDepthPats(d - 1, pat)
                        val restPats = genTuplePats(pats)
                    in
                        curPats :: restPats
                    end

                val generatedPatsLists = genTuplePats(tList)

                val patternSet = tupleCrossProduct(generatedPatsLists)

                fun tupleizeList(nil) = nil
                  | tupleizeList(aPatList::patLists) =
                    let
                        val curTuplePat = TuplePatternIntern(aPatList)
                        val restTuplePats = tupleizeList(patLists)
                    in
                        curTuplePat :: restTuplePats
                    end
            in
                tupleizeList(patternSet)
            end
          | generateDepthPats(d, RecordType(recs)) =
            let
                fun genRecPats(nil):{ident:string, pattern:PatternIntern} list list = nil
                  | genRecPats({tident=tid, ttype=tt}::recTypes) =
                    let
                        fun prependIdentToPatList(nil) =
                                nil:{ident:string, pattern:PatternIntern} list
                          | prependIdentToPatList(curPat::restPats) =
                            let
                                val curRec = {ident=tid, pattern=curPat}
                                val restRecs = prependIdentToPatList(restPats)
                            in
                                curRec :: restRecs
                            end

                        val curPats:PatternIntern list = generateDepthPats(d - 1, tt)
                        val curRecs:{ident:string, pattern:PatternIntern} list =
                                prependIdentToPatList(curPats)
                        val restRecs = genRecPats(recTypes)
                    in
                        curRecs :: restRecs
                    end

                val generatedRecsLists:{ident:string, pattern:PatternIntern} list list =
                        genRecPats(recs)

                fun debugPrintRecLists(nil:{ident:string, pattern:PatternIntern} list list) = print("\n")
                  | debugPrintRecLists(aRecList::restRecsLists) =
                    let
                        val _ = print("This paramter has the following patterns: ")
                        fun printRecList(nil:{ident:string, pattern:PatternIntern} list) = print("\n")
                          | printRecList({ident=i, pattern=p} :: restRecs) =
                            let
                                val _ = print("id = " ^ i ^ " pat = " ^ printPatternIntern(p) ^ ", ")
                            in
                                printRecList(restRecs)
                            end
                        val _ = printRecList(aRecList)
                    in
                        debugPrintRecLists(restRecsLists)
                    end
(*
                val _ = print("DEBUG: starting uncrossed printout\n")
                val _ = debugPrintRecLists(generatedRecsLists)
                val _ = print("DEBUG: finished uncrossed printout\n")
*)

                val recordSet = recordCrossProduct(generatedRecsLists)
(*
                val _ = print("DEBUG: starting cross producted printout\n")
                val _ = debugPrintRecLists(recordSet)
                val _ = print("DEBUG: finished cross producted printout\n")
*)

                fun recordizeList(nil) = nil
                  | recordizeList(aRecList::recLists) =
                    let
                        val curRecPat = RecordPatternIntern(aRecList)
                        val restRecPats = recordizeList(recLists)
                    in
                        curRecPat :: restRecPats
                    end
            in
                recordizeList(recordSet)
            end

(*
 * COLIN: ClassInstaceType seems to be quite popular... Still ignoring owner.
 *)
          | generateDepthPats(d, ClassInstanceType{class=cl, is_owner=owner}) =
            let
(*
 *              First get the concrete classes of our class.  We do this
 *              because we want to generate patterns for only concrete
 *              classes (can't instantiate instances of abstract classes).
 *              The function topConcreteClasses is defined below.
 *)
                val classSet:Class list = topConcreteClasses(cl, name_env)

(*
 *              genClassPats(): Given a class it generates all the possible
 *                  patterns for that class.
 *
 *              Parameters: a class (assumed to be concrete)
 *
 *              Return Value: a list of patterns (ClassPatternInterns) that can
 *                 be generated to fit the type of the given class.
 *)
                fun genClassPats(concreteClass) =
                    let
(*
 *                      Get the type of the class and use it to generate
 *                      patterns that match that type.
 *)
                        val clRep = getClassType(concreteClass)
                        val pats:PatternIntern list =
                           generateDepthPats(d - 1, clRep)

                        val _ = print("DEBUG: Generate the following patterns "
                            ^ "for class: " ^ printClassName(concreteClass) ^
                            " with type: " ^ printType(clRep) ^ "\n")
                        val _ = print(printPatternInternList(pats) ^ "\n")

(*
 *                      createClassPats(): creates a list of
 *                          ClassInternPatterns.
 *
 *                      Parameters: a list of patterns which are the
 *                          representation of the classes type.
 *
 *                      Return Value: a list of ClassPatternInterns using
 *                          the given patterns.
 *)
                        fun createClassPats(nil:PatternIntern list) =
                                nil:PatternIntern list
                          | createClassPats(hdPat::restPats) =
                            let
                                val classPat = ClassPatternIntern{
                                        class=concreteClass, rep=hdPat}
                                val restClassPats = createClassPats(restPats)
                            in
                                classPat :: restClassPats
                            end
                    in
                        createClassPats(pats)
                    end
                
                fun genClassesPats(nil:Class list) = nil:PatternIntern list
                  | genClassesPats(hdClass :: restClasses) =
                    let
                        val hdClassPattern:PatternIntern list = genClassPats(hdClass)
                        val tlClassPatterns:PatternIntern list=
                                genClassesPats(restClasses)
                    in
                        hdClassPattern @ tlClassPatterns
                    end
            in
                genClassesPats(classSet)
            end
          | generateDepthPats(_, t) = raise TypeError("Type: " ^ printType(t) ^
                " is unsupported!")
    in
        generateDepthPats(patDepth, argType)
    end

and topConcreteClasses((cl as Class{name=cName, id=nID, mode=cMode,
        superclass=cSuper, interfaces=cInters, dType=cType}), name_env) =
    let
(*
 *      topConClassList: given a list of classes, generate the top concrete
 *          classes for each element in the list.
 *)
        fun topConClassList(nil:Class list) = nil:Class list
          | topConClassList(hdCl::tlCls) = 
            let
                val hdConcClasses:Class list = topConcreteClasses(hdCl,
                        name_env)
                val tlConcClasses:Class list = topConClassList(tlCls)
            in
                hdConcClasses @ tlConcClasses
            end
    in
        case (cMode) of
            Concrete => [cl]
          | Interface =>
(*
 *          To find an interface's concrete classes, we must find all classes that
 *          implement that interface.
 *)
            let
                val implementingClasses = getImplementingClasses(cl, name_env)
            in
                topConClassList(implementingClasses)
            end
          | Abstract =>
            let
                val directChildClasses = getClassChildren(cl, name_env)
            in
                topConClassList(directChildClasses)
            end
    end

and tupleCrossProduct(nil):PatternIntern list list = [nil]
  | tupleCrossProduct(downList :: acrossLists) =
    let
        val restProduct:PatternIntern list list = tupleCrossProduct(acrossLists)

        fun prependEltToSet(elt:PatternIntern, nil:PatternIntern list list):
                PatternIntern list list = (*(elt :: nil:PatternIntern list) ::*) (* <colin> mirror comments below?*)
                    nil:PatternIntern list list
          | prependEltToSet(elt, setHead :: setRest) =
            let
                val newList:PatternIntern list = elt :: setHead
                val restList:PatternIntern list list = prependEltToSet(elt, setRest)
            in
                newList :: restList
            end

        fun prependDownListToSet(nil) = nil
          | prependDownListToSet(downElt :: restDown) =
            let
                val restPrepends = prependDownListToSet(restDown)
                val curPrepends = prependEltToSet(downElt, restProduct)
            in
                curPrepends @ restPrepends
            end
    in
        prependDownListToSet(downList)
    end

(*
 *  recordCrossProduct(): Computes the set of record lists that can be created
 *      from the given input.
 *
 *  Parameters: A list of lists of {ident:string, pattern:PatternIntern} (the
 *      internal representation of RecordPatternIntern).
 *
 *  Return value: A list of lists that is the cross product of the input list
 *      lists.
 *)

and recordCrossProduct(nil):{ident:string, pattern:PatternIntern} list list =
        [nil]
  | recordCrossProduct(downList :: acrossLists) =
    let
        val restProduct:{ident:string, pattern:PatternIntern} list list =
                recordCrossProduct(acrossLists)



                        fun printRecList(nil:{ident:string, pattern:PatternIntern} list) = print("\n")
                          | printRecList({ident=i, pattern=p} :: restRecs) =
                            let
                                val _ = print("id = " ^ i ^ " pat = " ^ printPatternIntern(p) ^ ", ")
                            in
                                printRecList(restRecs)
                            end


                fun debugPrintRecLists(nil:{ident:string, pattern:PatternIntern} list list) = print("\n")
                  | debugPrintRecLists(aRecList::restRecsLists) =
                    let
                        val _ = print("This paramter has the following patterns: ")

                        val _ = printRecList(aRecList)
                    in
                        debugPrintRecLists(restRecsLists)
                    end
(*
        val _ = print("DEBUG: currently handling:\n")
        val _ = printRecList(downList)
        val _ = print("DEBUG: rest of product looks like:\n")
        val _ = debugPrintRecLists(restProduct)
        val _ = print("DEBUG: finished rest of product.\n")
*)
        fun prependEltToSet(elt:{ident:string, pattern:PatternIntern},
                            nil:{ident:string, pattern:PatternIntern} list list):
                            {ident:string, pattern:PatternIntern} list list =
                (*(elt :: nil:{ident:string, pattern:PatternIntern} list) ::*) (* <colin> it now works? *)
                    nil:{ident:string, pattern:PatternIntern} list list
          | prependEltToSet(elt, setHead :: setRest) =
            let
                val newList:{ident:string, pattern:PatternIntern} list = elt :: setHead
                val restList:{ident:string, pattern:PatternIntern} list list =
                        prependEltToSet(elt, setRest)
            in
                newList :: restList
            end

        fun prependDownListToSet(nil) = nil
          | prependDownListToSet(downElt :: restDown) =
            let
                val restPrepends = prependDownListToSet(restDown)
                val curPrepends = prependEltToSet(downElt, restProduct)
            in
                curPrepends @ restPrepends
            end
    in
        prependDownListToSet(downList)
    end

(*
 *  depthPattern(): get the depth of a pattern.
 *
 *  Parameters: the pattern
 *
 *  Return value: the depth of the specified pattern.
 *
 *  Comments: <colin> Unresolved issues about the depth of a nil pattern and an
 *      empty Tuple (or Record) pattern).
 *)
and depthPattern(VarPatternIntern(_)) = 0
  | depthPattern(WildcardPatternIntern) = 0
  | depthPattern(IntPatternIntern(_)) = 1
  | depthPattern(RealPatternIntern(_)) = 1
  | depthPattern(StringPatternIntern(_)) = 1
  | depthPattern(BoolPatternIntern(_)) = 1
  | depthPattern(NilPatternIntern) = 1
  | depthPattern(TuplePatternIntern(pats)) = 1 + depthPatterns(pats, 0)
  | depthPattern(RecordPatternIntern(recs)) = 1 + depthRecPatterns(recs, 0)
  | depthPattern(ConsPatternIntern{head, tail}) =
    let
        val headDepth = depthPattern(head)
        val tailDepth = depthPattern(tail)
    in
        if (headDepth > tailDepth) then
            1 + headDepth
        else
            1 + tailDepth
    end
  | depthPattern(ClassPatternIntern{class, rep}) = 1 + depthPattern(rep)

(*
 *  depthPatterns(): get the depth of a list of patterns.
 *
 *  Parameters: the list of patterns to check, the length of the previous
 *      pattern (should be initialized to 0).
 *
 *  Return value: the max of the depths of each pattern in the list.
 *
 *  Comments:
 *)
and depthPatterns(nil, prevDepth) = prevDepth
  | depthPatterns(curPat::restPats, prevDepth) =
    let
        val curDepth = depthPattern(curPat)
    in
        if (curDepth > prevDepth) then
            depthPatterns(restPats, curDepth)
        else
            depthPatterns(restPats, prevDepth)
    end

(*
 *  depthRecPatterns(): get the depth of a list of record patterns.
 *
 *  Parameters: the list of record patterns to check, the length of the
 *      previous pattern (should be initialized to 0).
 *
 *  Return value: the max of the depths of each pattern in the list.
 *
 *  Comments: Similar to depthPatterns.
 *)
and depthRecPatterns(nil, prevDepth) = prevDepth
  | depthRecPatterns({ident, pattern}::restRecs, prevDepth) =
    let
        val curDepth = depthPattern(pattern)
    in
        if (curDepth > prevDepth) then
            depthRecPatterns(restRecs, curDepth)
        else
            depthRecPatterns(restRecs, prevDepth)
    end

(*
 *  depthType(): Returns the depth of a type.
 *
 *  Parameters: The type to get the depth from.
 *
 *  Return value: Returns the depth of the type.
 *
 *  Comments: The depth of a Type is calculated based on Todd's work.
 *)
and depthType(IntegerType) = 1
  | depthType(RealType) = 1
  | depthType(BooleanType) = 1
  | depthType(StringType) = 1
  | depthType(TupleType(tList)) = 1 + depthTupleTypes(tList, 0)
  | depthType(ListType(t)) = 1 + depthType(t)
(*
 * COLIN: marking all places I ignore owner (#)
 *)
  | depthType(ClassInstanceType(_)) = 1
  | depthType(RecordType(recs)) = 1 + depthRecordTypes(recs, 0)
  | depthType(FunctionType(_, _)) = 0
  | depthType(UnevaluatedClassType(_)) = raise TypeError("Unevaluated class " ^
        "type should have NEVER made it here!  Bailing!")

(*
 *  depthTupleTypes(): Calculates the depth of a tuple type.
 *
 *  Parameters: A list of Types (which is what a TupleType consists of) and the
 *      largest depth of the records so far included in the parameters for
 *      simplicity only, should be passed in as 0!)
 *
 *  Return value: Returns the maximum of the depth of each elements in the
 *      list.
 *
 *  Comments: If the list is empty the previous maximum is returned.  Otherwise
 *      depth of the current Type is determined and compared with the previous
 *      maximum.  If the new max is greater, recursively call depthTupleTypes()
 *      with the new maximum (otherwise continue with the previous).  The
 *      prevDepth is kept in the function prototype for programmer convenience
 *      (recursion).
 *)
and depthTupleTypes(nil, prevDepth) = prevDepth
  | depthTupleTypes(curType::restTypes, prevDepth) =
    let
        val curDepth = depthType(curType)
    in
        if (curDepth > prevDepth) then
            depthTupleTypes(restTypes, curDepth)
        else
            depthTupleTypes(restTypes, prevDepth)
    end

(*
 *  depthRecordTypes(): Calculates the depth of a record type.
 *
 *  Parameters: A list of {tident:string, ttype:Type} (which is what a record
 *      type consists of) and the largest depth of the records so far
 *      included in the parameters for simplicity only, should be passed in as
 *      0!)
 *
 *  Return value: Returns the maximum of the depth of each elements in the
 *      list.
 *
 *  Comments: Similar to depthTupleTypes() code.  The prevDepth is kept in the
 *      function prototype for programmer convenience (recursion).
 *)
and depthRecordTypes(nil, prevDepth) = prevDepth
  | depthRecordTypes({tident, ttype}::restTypes, prevDepth) =
    let
        val curDepth = depthType(ttype)
    in
        if (curDepth > prevDepth) then
            depthRecordTypes(restTypes, curDepth)
        else
            depthRecordTypes(restTypes, prevDepth)
    end

(*
 *  findOwnerAmbiguities(): Given the global type environment (the type
 *      environment before the new bindings are added) and the local type
 *      environment (the new bindings), ensure that there are no ambiguities.
 *      This is done by:
 *          - Loop through each method
 *          - For each method, check to see if the message was declared in
 *              the global scope.
 *          - If it was not declared in the global scope, then ignore the
 *              current method        
 *          - If it was foud in the global scope, then ensure that the
 *              method dispatches on a LOCAL class in the owner position.
 *              That is, make sure that the class specified in the owner
 *              position is also declared locally.
 *
 *  Parameters: A global scope, a local scope, and a list of local methods that
 *      were just created (they aren't included in the local scope, so we need
 *      to explicitly get them).  See function description for more details.
 *
 *  Return Value: Returns 0 if there are no ambiguities; raises a TypeError if
 *      any are found.  So, if this function returns normally then everything
 *      okay.
 *
 *  Comments: This is the first of three restrictions imposed to emulate
 *      modular type checking.
 *)
and findOwnerAmbiguities(global_tenv, local_tenv, method_env) =

(*
 * Colin: for now just return, code in progress has been commented out.
 *)
    ()
(*
    let

        fun checkLocalMethods(nil) =
            (*
             *  If the input is empty then we've successfully searched all the
             *  local methods.  Print out an informational message and return
             *  success (0).
             *)
            let
                val _ = emlPrint("DEBUG: no more local methods to check!\n")
            in
                0
            end
          | checkLocalMethods((NONE) :: rest_methods) =
            (*
             *  For simplicity in they initial type checker, non method
             *  declarations returned NONE, so we filter those out here.
             *)
            let
                val _ = emlPrint("DEBUG: encountered NONE method; ignoring!\n")
            in
                checkLocalMethods(rest_methods)
            end
          | checkLocalMethods((SOME({name, formal, body})) :: rest_methods) =
            (*
             *  Check the current method to see if its message was declared
             *  globally.  If it was, then ensure that the owner class the
             *  method dispatches on is locally declared.
             *)
            let
                val _ = print("DEBUG: checking local method: " ^ name ^ "\n")

                (*
                 *  Check the global environment to see if the message exists
                 *  there.  If it does not (NotFound is returned) then we can
                 *  move onto the next local method.  We do this by setting
                 *  global type to NONE, which will make the rest of the step
                 *  just ignore the current value and we'll move onto the next.
                 *
                 *  If the message is found then the type is returned, and we
                 *  can use that type to verify that we're comparing the method
                 *  to the correct message.
                 *)

               val global_type = lookup_tbinding(global_tenv, name)
                    handle NotFound => checkLocalMethods(rest_methods)

                val owner_class = 
                    if (global_type = SOME(_)) then
                        SOME(findOwnerClassName(actual_type))
                            handle NotFound => raise TypeError("No owner " ^
                                "class found for message " ^ name ^ "!")
                    else
                        NONE

                (*
                 *  If we successfully find the class name in the local
                 *  environment then we are okay.  If it is not found
                 *  then we raise a type error.
                 *)
                val class_is_local = 
                    if (owner_class = SOME(owner_name)) then
                        SOME(lookup_tbinding(local_tenv, owner_name))
                        handle NotFound => raise TypeError("Local method " ^
                            name ^ " must dipatch on  a local class in the " ^
                            " owner position!")
                    else
                        NONE
                    
            in
                (*
                 *  If we've made it this far then the current method is fine,
                 *  check the rest of them.
                 *)
                checkLocalMethods(rest_methods)
            end
    in
        checkLocalMethods(method_env)
    end
*)
